#! /usr/bin/env python
import pytest

import numpy as np
from numpy.testing import assert_array_equal

from landlab.field.graph_field import GraphFields as ModelDataFields
from landlab.field import GroupError, FieldError


def test_init():
    fields = ModelDataFields()
    assert set() == fields.groups
    # assert_set_equal(set(), fields.groups)


def test_new_field_location():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    assert set(['node']) == fields.groups
    # assert_set_equal(set(['node']), fields.groups)


def test_add_existing_group():
    fields = ModelDataFields()
    fields.new_field_location('node', size=12)
    with pytest.raises(ValueError):
        fields.new_field_location('node', size=24)


def test_add_multiple_groups():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.new_field_location('cell', 2)
    fields.new_field_location('face', 7)
    fields.new_field_location('link', 7)
    assert set(['node', 'cell', 'face', 'link']) == fields.groups
    # assert_set_equal(set(['node', 'cell', 'face', 'link']), fields.groups)


def test_ones():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.new_field_location('cell', 2)

    value_array = fields.ones('node')
    assert_array_equal(np.ones(12), value_array)

    value_array = fields.ones('cell')
    assert_array_equal(np.ones(2), value_array)


def test_add_ones():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.new_field_location('cell', 2)

    fields.add_ones('z', at='node')
    assert_array_equal(np.ones(12), fields['node']['z'])
    assert_array_equal(np.ones(12), fields.field_values('node', 'z'))

    fields.add_ones('z', at='cell')
    assert_array_equal(np.ones(2), fields['cell']['z'])
    assert_array_equal(np.ones(2), fields.field_values('cell', 'z'))


def test_add_ones_return_value():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.new_field_location('cell', 2)

    rtn_value = fields.add_ones('z', at='node')
    assert_array_equal(rtn_value, np.ones(12))
    assert rtn_value is fields['node']['z']
    assert rtn_value is fields.field_values('node', 'z')

    rtn_value = fields.add_ones('z', at='cell')
    assert_array_equal(rtn_value, np.ones(2))
    assert rtn_value is fields['cell']['z']
    assert rtn_value is fields.field_values('cell', 'z')


def test_add_existing_field_default():
    """Test default is to not replace existing field."""
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.add_empty('z', at='node')

    with pytest.raises(FieldError):
        fields.add_empty('z', at='node')
    with pytest.raises(FieldError):
        fields.add_ones('z', at='node')
    with pytest.raises(FieldError):
        fields.add_zeros('z', at='node')


def test_add_existing_field_with_noclobber():
    """Test noclobber raises an error with an existing field."""
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.add_empty('z', at='node')

    with pytest.raises(FieldError):
        fields.add_empty('z', at='node', noclobber=True)
    with pytest.raises(FieldError):
        fields.add_ones('z', at='node', noclobber=True)
    with pytest.raises(FieldError):
        fields.add_zeros('z', at='node', noclobber=True)


def test_add_field_with_noclobber():
    """Test noclobber does not raise an error with an new field."""
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    fields.add_empty('a', at='node', noclobber=True)
    assert 'a' in fields['node']

    fields.add_ones('b', at='node', noclobber=True)
    assert 'b' in fields['node']

    fields.add_zeros('c', at='node', noclobber=True)
    assert 'c' in fields['node']


def test_add_field_with_clobber():
    """Test adding a field with clobber on."""
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    assert fields.add_empty('a', at='node') is not fields.add_empty('a', at='node', noclobber=False)
    assert fields.add_ones('b', at='node') is not fields.add_ones('b', at='node', noclobber=False)
    assert fields.add_zeros('c', at='node') is not fields.add_zeros('c', at='node', noclobber=False)


def test_getitem():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    assert dict() == fields['node']
    with pytest.raises(GroupError):
        fields['cell']
    with pytest.raises(KeyError):
        fields['cell']


def test_at_attribute():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    assert dict() == fields.at_node
    with pytest.raises(AttributeError):
        fields.at_cell

    fields.add_ones('z', at='node')
    assert_array_equal(np.ones(12), fields.at_node['z'])


def test_has_group():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    assert fields.has_group('node')
    assert not fields.has_group('cell')


def test_delete_field():
    fields = ModelDataFields()
    fields.new_field_location('link', 17)

    assert dict() == fields.at_link
    with pytest.raises(AttributeError):
        fields.at_node

    fields.add_zeros('link', 'vals')
    assert_array_equal(np.zeros(17), fields.at_link['vals'])

    with pytest.raises(KeyError):
        fields.delete_field('node', 'vals')
    fields.delete_field('link', 'vals')
    with pytest.raises(KeyError):
        fields.field_units('link', 'vals')
    with pytest.raises(KeyError):
        fields.at_link['vals']


def test_scalar_field():
    """Test adding a generic scalar field."""
    fields = ModelDataFields()
    fields.new_field_location('all_over_the_place', 1)

    assert dict() == fields.at_all_over_the_place
    with pytest.raises(AttributeError):
        fields.at_cell

    fields.at_all_over_the_place['const'] = 1.
    assert_array_equal(np.array(1.), fields.at_all_over_the_place['const'])
    
    val = np.array(2.)
    fields.at_all_over_the_place['const'] = val
    assert val is fields.at_all_over_the_place['const']


def test_grid_field_as_array():
    """Test adding an array as a grid field."""
    fields = ModelDataFields()
    fields.new_field_location('grid', 1)

    fields.at_grid['const'] = [1., 2.]
    assert_array_equal(np.array([1., 2.]), fields.at_grid['const'])

    val = np.array([1., 2.])
    fields.at_grid['const'] = val
    assert val is fields.at_grid['const']

    val.shape = (1, 1, 2, 1)
    fields.at_grid['const'] = val
    assert_array_equal(np.array([1., 2.]), fields.at_grid['const'])
    assert val is fields.at_grid['const']


def test_grid_field_add_zeros_ones_empty():
    """Test creating scalar fields with add_zeros, add_empty, and add_ones."""
    fields = ModelDataFields()
    fields.new_field_location('grid', 1)
    
    with pytest.raises(ValueError):
        fields.add_zeros('value', at='grid')
    with pytest.raises(ValueError):
        fields.add_empty('value', at='grid')
    with pytest.raises(ValueError):
        fields.add_ones('value', at='grid')

def test_grid_field_zeros_ones_empty():
    """Test creating scalar fields with zeros, empty, and ones."""
    fields = ModelDataFields()
    fields.new_field_location('grid', 1)    
    with pytest.raises(ValueError):
        fields.zeros('grid')
    with pytest.raises(ValueError):
        fields.empty('grid')
    with pytest.raises(ValueError):
        fields.ones('grid')

def test_nd_field():
    """Test creating fields that are nd in shape."""
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    fields.add_field('new_value', np.ones((12,4,5)), at='node')
    fields.add_field('newer_value', np.ones((12,4)), at='node')

    with pytest.raises(ValueError):
        fields.add_field('newest_value', np.ones((13,4,5)), at='node')
    with pytest.raises(ValueError):
        fields.add_field('newestest_value', np.ones((13)), at='node')

    



# -*- coding: utf-8 -*-
"""
Unit tests for landlab.components.overland_flow.KinwaveOverlandFlowModel

last updated: 3/14/16
"""
import numpy as np

from landlab import RasterModelGrid
from landlab.components.overland_flow import KinwaveOverlandFlowModel


(_SHAPE, _SPACING, _ORIGIN) = ((10, 10), (25, 25), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_KinWaveOF_name(kin_wave_of):
    assert kin_wave_of.name == 'KinwaveOverlandFlowModel'


def test_KinWaveOF_input_var_names(kin_wave_of):
    assert kin_wave_of.input_var_names == (
        'topographic__elevation',
        'topographic__gradient',
    )


def test_KinWaveOF_output_var_names(kin_wave_of):
    assert kin_wave_of.output_var_names == (
        'surface_water__depth',
        'water__velocity',
        'water__specific_discharge',
    )


def test_KinWaveOF_var_units(kin_wave_of):
    assert set(kin_wave_of.input_var_names) | set(kin_wave_of.output_var_names) == set(dict(kin_wave_of.units).keys())

    assert kin_wave_of.var_units('topographic__elevation') == 'm'
    assert kin_wave_of.var_units('topographic__gradient') == 'm/m'
    assert kin_wave_of.var_units('surface_water__depth') == 'm'
    assert kin_wave_of.var_units('water__velocity') == 'm/s'
    assert kin_wave_of.var_units('water__specific_discharge') == 'm2/s'


def test_grid_shape(kin_wave_of):
    assert kin_wave_of.grid.number_of_node_rows == _SHAPE[0]
    assert kin_wave_of.grid.number_of_node_columns == _SHAPE[1]


def test_run_one_step():
    from landlab import RasterModelGrid
    import numpy as np
    from landlab.components.overland_flow import KinwaveOverlandFlowModel

    grid = RasterModelGrid((10, 10), spacing=0.5)
    grid.add_zeros('node', 'topographic__elevation', dtype=float)
    grid.add_zeros('node', 'topographic__gradient')

    topo_arr = np.ones(100).reshape(10, 10)
    i=0
    while i <= 9:
        topo_arr[:, i]  = 5 + (0.002*i)
        i+=1
    topo_arr = topo_arr.flatten()
    grid['node']['topographic__elevation'] = topo_arr
    KinWaveOF = KinwaveOverlandFlowModel(grid, precip_rate=100.,
        precip_duration=1.0, roughness=0.02)

    KinWaveOF.run_one_step(60)

    # I'll admit this is very non-robust. Solution roughly based on plot #9
    # from Heng et. al, (2009): "Modeling overland flow and soil eroion on
    # non uniform hillslopes: A finite volume scheme." They do not provide the
    # numerical solution but the plots match...
    max_h_mm = max(grid['node']['surface_water__depth']) * 1000.
    np.testing.assert_almost_equal(max_h_mm, 1.66666666667)



#! /usr/bin/env python
"""
Defines the base component class from which Landlab components inherit.

Base component class methods
++++++++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.core.model_component.Component.from_path
    ~landlab.core.model_component.Component.name
    ~landlab.core.model_component.Component.units
    ~landlab.core.model_component.Component.definitions
    ~landlab.core.model_component.Component.input_var_names
    ~landlab.core.model_component.Component.output_var_names
    ~landlab.core.model_component.Component.optional_var_names
    ~landlab.core.model_component.Component.var_type
    ~landlab.core.model_component.Component.var_units
    ~landlab.core.model_component.Component.var_definition
    ~landlab.core.model_component.Component.var_mapping
    ~landlab.core.model_component.Component.var_loc
    ~landlab.core.model_component.Component.var_help
    ~landlab.core.model_component.Component.initialize_output_fields
    ~landlab.core.model_component.Component.initialize_optional_output_fields
    ~landlab.core.model_component.Component.shape
    ~landlab.core.model_component.Component.grid
    ~landlab.core.model_component.Component.coords
    ~landlab.core.model_component.Component.imshow
"""

from __future__ import print_function

import os
import textwrap
import warnings
import inspect

from .. import registry

_VAR_HELP_MESSAGE = """
name: {name}
description:
{desc}
units: {units}
at: {loc}
intent: {intent}
"""


class classproperty(property):
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


class Component(object):
    """
    Defines the base component class from which Landlab components inherit.

    **Base component class methods**

    .. autosummary::
        :toctree: generated/

        ~landlab.core.model_component.Component.from_path
        ~landlab.core.model_component.Component.name
        ~landlab.core.model_component.Component.units
        ~landlab.core.model_component.Component.definitions
        ~landlab.core.model_component.Component.input_var_names
        ~landlab.core.model_component.Component.output_var_names
        ~landlab.core.model_component.Component.optional_var_names
        ~landlab.core.model_component.Component.var_type
        ~landlab.core.model_component.Component.var_units
        ~landlab.core.model_component.Component.var_definition
        ~landlab.core.model_component.Component.var_mapping
        ~landlab.core.model_component.Component.var_loc
        ~landlab.core.model_component.Component.var_help
        ~landlab.core.model_component.Component.initialize_output_fields
        ~landlab.core.model_component.Component.initialize_optional_output_fields
        ~landlab.core.model_component.Component.shape
        ~landlab.core.model_component.Component.grid
        ~landlab.core.model_component.Component.coords
        ~landlab.core.model_component.Component.imshow
    """
    _input_var_names = set()
    _output_var_names = set()
    _optional_var_names = set()
    _var_units = dict()

    def __new__(cls, *args, **kwds):
        registry.add(cls)
        return object.__new__(cls)

    def __init__(self, grid, map_vars=None, **kwds):
        map_vars = map_vars or {}
        self._grid = grid

        for (location, vars) in map_vars.items():
            for (dest, src) in vars.items():
                grid.add_field(location, dest,
                               grid.field_values(location, src))

        for key in kwds:
            component_name = inspect.getmro(self.__class__)[0].__name__
            warnings.warn(
                "Ingnoring unrecognized input parameter, '{param}', for "
                "{name} component".format(name=component_name, param=key))

    @classmethod
    def from_path(cls, grid, path):
        """Create a component from an input file.

        Parameters
        ----------
        grid : ModelGrid
            A landlab grid.
        path : str or file_like
            Path to a parameter file, contents of a parameter file, or
            a file-like object.

        Returns
        -------
        Component
            A newly-created component.
        """
        if os.path.isfile(path):
            with open(path, 'r') as fp:
                params = load_params(fp)
        else:
            params = load_params(path)
        return cls(grid, **params)

    @classproperty
    @classmethod
    def input_var_names(cls):
        """Names of fields that are used by the component.

        Returns
        -------
        tuple of str
            Tuple of field names.
        """
        return tuple(cls._input_var_names)

    @classproperty
    @classmethod
    def output_var_names(self):
        """Names of fields that are provided by the component.

        Returns
        -------
        tuple of str
            Tuple of field names.
        """
        return tuple(self._output_var_names)

    @classproperty
    @classmethod
    def optional_var_names(self):
        """
        Names of fields that are optionally provided by the component, if
        any.

        Returns
        -------
        tuple of str
            Tuple of field names.
        """
        try:
            return tuple(self._optional_var_names)
        except AttributeError:
            return ()

    @classmethod
    def var_type(cls, name):
        """
        Returns the dtype of a field (float, int, bool, str...), if declared.
        Default is float.

        Parameters
        ----------
        name : str
            A field name.

        Returns
        -------
        dtype
            The dtype of the field.
        """
        try:
            return cls._var_type[name]
        except AttributeError:
            return float

    @classproperty
    @classmethod
    def name(self):
        """Name of the component.

        Returns
        -------
        str
            Component name.
        """
        return self._name

    @classproperty
    @classmethod
    def units(self):
        """Get the units for all field values.

        Returns
        -------
        tuple or str
            Units for each field.
        """
        return tuple(self._var_units.items())

    @classmethod
    def var_units(cls, name):
        """Get the units of a particular field.

        Parameters
        ----------
        name : str
            A field name.

        Returns
        -------
        str
            Units for the given field.
        """
        return cls._var_units[name]

    @classproperty
    @classmethod
    def definitions(cls):
        """Get a description of each field.

        Returns
        -------
        tuple of (*name*, *description*)
            A description of each field.
        """
        return tuple(cls._var_doc.items())

    @classmethod
    def var_definition(cls, name):
        """Get a description of a particular field.

        Parameters
        ----------
        name : str
            A field name.

        Returns
        -------
        tuple of (*name*, *description*)
            A description of each field.
        """
        return cls._var_doc[name]

    @classmethod
    def var_help(cls, name):
        """Print a help message for a particular field.

        Parameters
        ----------
        name : str
            A field name.
        """
        desc = os.linesep.join(textwrap.wrap(cls._var_doc[name],
                                             initial_indent='  ',
                                             subsequent_indent='  '))
        units = cls._var_units[name]
        loc = cls._var_mapping[name]

        intent = ''
        if name in cls._input_var_names:
            intent = 'in'
        if name in cls._output_var_names:
            intent += 'out'

        help = _VAR_HELP_MESSAGE.format(name=name, desc=desc, units=units,
                                        loc=loc, intent=intent)

        print(help.strip())

    @classproperty
    @classmethod
    def var_mapping(self):
        """Location where variables are defined.

        Returns
        -------
        tuple of (name, location)
            Tuple of variable name and location ('node', 'link', etc.) pairs.
        """
        return tuple(self._var_mapping.items())

    @classmethod
    def var_loc(cls, name):
        """Location where a particular variable is defined.

        Parameters
        ----------
        name : str
            A field name.

        Returns
        -------
        str
            The location ('node', 'link', etc.) where a variable is defined.
        """
        return cls._var_mapping[name]

    def initialize_output_fields(self):
        """
        Create fields for a component based on its input and output var names.

        This method will create new fields (without overwrite) for any fields
        output by, but not supplied to, the component. New fields are
        initialized to zero. Ignores optional fields, if specified by
        _optional_var_names. New fields are created as arrays of floats, unless
        the component also contains the specifying property _var_type.
        """
        for field_to_set in (set(self.output_var_names) -
                             set(self.input_var_names) -
                             set(self.optional_var_names)):
            grp = self.var_loc(field_to_set)
            type_in = self.var_type(field_to_set)
            init_vals = self.grid.zeros(grp, dtype=type_in)
            units_in = self.var_units(field_to_set)
            self.grid.add_field(grp,
                                field_to_set,
                                init_vals,
                                units=units_in,
                                copy=False,
                                noclobber=True)

    def initialize_optional_output_fields(self):
        """
        Create fields for a component based on its optional field outputs,
        if declared in _optional_var_names.

        This method will create new fields (without overwrite) for any fields
        output by the component as optional. New fields are
        initialized to zero. New fields are created as arrays of floats, unless
        the component also contains the specifying property _var_type.
        """
        for field_to_set in (set(self.optional_var_names) -
                             set(self.input_var_names)):
            self.grid.add_field(self.var_loc(field_to_set),
                                field_to_set,
                                self.grid.zeros(
                                    dtype=self.var_type(field_to_set)),
                                units=self.var_units(field_to_set),
                                noclobber=True)

    @property
    def shape(self):
        """Return the grid shape attached to the component, if defined."""
        return self.grid._shape

    @property
    def grid(self):
        """Return the grid attached to the component."""
        return self._grid

    @property
    def coords(self):
        """Return the coordinates of nodes on grid attached to the component.
        """
        return (self.grid.node_x, self.grid.node_y)

    def imshow(self, name, **kwds):
        """Plot data on the grid attached to the component.
        """
        self._grid.imshow(name, **kwds)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  3 10:39:32 2017

@author: gtucker
"""
import pytest
import warnings

import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.components import DepthDependentDiffuser, ExponentialWeatherer


def test_raise_kwargs_error():
    mg = RasterModelGrid((5, 5))
    soilTh = mg.add_zeros('node', 'soil__depth')
    z = mg.add_zeros('node', 'topographic__elevation')
    BRz = mg.add_zeros('node', 'bedrock__elevation')
    z += mg.node_x.copy()
    BRz += mg.node_x/2.
    soilTh[:] = z - BRz
    expweath = ExponentialWeatherer(mg)
    with pytest.raises(TypeError):
        DepthDependentDiffuser(mg, diffusivity=1)
    
    DDdiff = DepthDependentDiffuser(mg)
    
    with pytest.raises(TypeError):
        DDdiff.soilflux(2., bad_var=1)



from .graph import Graph
from .dual import DualGraph
from .structured_quad import *
from .voronoi import *
from .hex import *
from .radial import *


__all__ = ['Graph', 'DualGraph', 'StructuredQuadGraph', 'RectilinearGraph',
           'UniformRectilinearGraph', 'DualUniformRectilinearGraph',
           'DualRectilinearGraph', 'DualStructuredQuadGraph', 'VoronoiGraph',
           'DualVoronoiGraph', 'HexGraph', 'DualHexGraph', 'RadialGraph',
           'DualRadialGraph', ]



"""Simulate detachment limited sediment transport.

Landlab component that simulates detachment limited sediment transport is more
general than the stream power component. Doesn't require the upstream node
order, links to flow receiver and flow receiver fields. Instead, takes in
the discharge values on NODES calculated by the OverlandFlow class and
erodes the landscape in response to the output discharge.

As of right now, this component relies on the OverlandFlow component 
for stability. There are no stability criteria implemented in this class. 
To ensure model stability, use StreamPowerEroder or FastscapeEroder
components instead. 

.. codeauthor:: Jordan Adams

Examples
--------
>>> import numpy as np
>>> from landlab import RasterModelGrid
>>> from landlab.components import DetachmentLtdErosion

Create a grid on which to calculate detachment ltd sediment transport.

>>> grid = RasterModelGrid((4, 5))

The grid will need some data to provide the detachment limited sediment
transport component. To check the names of the fields that provide input to
the detachment ltd transport component, use the *input_var_names* class
property.

Create fields of data for each of these input variables.

>>> grid.at_node['topographic__elevation'] = np.array([
...     0., 0., 0., 0., 0.,
...     1., 1., 1., 1., 1.,
...     2., 2., 2., 2., 2.,
...     3., 3., 3., 3., 3.])

Using the set topography, now we will calculate slopes on all nodes.


>>> grid.at_node['topographic__slope'] = np.array([
...     -0.        , -0.        , -0.        , -0.        , -0,
...      0.70710678,  1.        ,  1.        ,  1.        ,  0.70710678,
...      0.70710678,  1.        ,  1.        ,  1.        ,  0.70710678,
...      0.70710678,  1.        ,  1.        ,  1.        ,  0.70710678])


Now we will arbitrarily add water discharge to each node for simplicity.
>>> grid.at_node['surface_water__discharge'] = np.array([
...     30., 30., 30., 30., 30.,
...     20., 20., 20., 20., 20.,
...     10., 10., 10., 10., 10.,
...      5., 5., 5., 5., 5.])

Instantiate the `DetachmentLtdErosion` component to work on this grid, and
run it. In this simple case, we need to pass it a time step ('dt')

>>> dt = 10.0
>>> dle = DetachmentLtdErosion(grid)
>>> dle.erode(dt=dt)

After calculating the erosion rate, the elevation field is updated in the
grid. Use the *output_var_names* property to see the names of the fields that
have been changed.

>>> dle.output_var_names
('topographic__elevation',)

The `topographic__elevation` field is defined at nodes.

>>> dle.var_loc('topographic__elevation')
'node'


Now we test to see how the topography changed as a function of the erosion
rate.

>>> grid.at_node['topographic__elevation'] # doctest: +NORMALIZE_WHITESPACE
array([ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
        0.99936754,  0.99910557,  0.99910557,  0.99910557,  0.99936754,
        1.99955279,  1.99936754,  1.99936754,  1.99936754,  1.99955279,
        2.99968377,  2.99955279,  2.99955279,  2.99955279,  2.99968377])

"""

from landlab import Component
import pylab
import numpy as np
from matplotlib import pyplot as plt
from landlab.field.scalar_data_fields import FieldError


class DetachmentLtdErosion(Component):

    """Landlab component that simulates detachment-limited river erosion.

    This component calculates changes in elevation in response to vertical
    incision.
    """

    _name = 'DetachmentLtdErosion'

    _input_var_names = (
        'topographic__elevation',
        'topographic__slope',
        'surface_water__discharge',
    )

    _output_var_names = (
        'topographic__elevation',
    )

    _var_units = {
        'topographic__elevation': 'm',
        'topographic__slope': '-',
        'surface_water__discharge': 'm^3/s',
    }

    _var_mapping = {
        'topographic__elevation': 'node',
        'topographic__slope': 'node',
        'surface_water__discharge': 'node',
    }

    _var_doc = {
        'topographic__elevation': 'Land surface topographic elevation',
        'topographic__slope': 'Slope of ',
        'surface_water__discharge': 'node',
    }

    def __init__(self, grid, K_sp = 0.00002, m_sp = 0.5, n_sp = 1.0,
                 uplift_rate = 0.0, entrainment_threshold = 0.0, **kwds):
        """Calculate detachment limited erosion rate on nodes.

        Landlab component that generalizes the detachment limited erosion
        equation, primarily to be coupled to the the Landlab OverlandFlow
        component.

        This component adjusts topographic elevation and is contained in the
        landlab.components.detachment_ltd_sed_trp folder.

        Parameters
        ----------
        grid : RasterModelGrid
            A landlab grid.
        K_sp : float, optional
            K in the stream power equation (units vary with other parameters -
            if used with the de Almeida equation it is paramount to make sure
            the time component is set to *seconds*, not *years*!)
        m_sp : float, optional
            Stream power exponent, power on discharge
        n_sp : float, optional
            Stream power exponent, power on slope
        uplift_rate : float, optional
            changes in topographic elevation due to tectonic uplift
        entrainment_threshold : float, optional
            threshold for sediment movement
        """
        super(DetachmentLtdErosion, self).__init__(grid, **kwds)

        self.K = K_sp
        self.m = m_sp
        self.n = n_sp

        self.I = self._grid.zeros(at='node')
        self.uplift_rate = uplift_rate
        self.entrainment_threshold = entrainment_threshold

        self.dzdt = self._grid.zeros(at='node')

    def erode(self, dt, elevs='topographic__elevation', 
              discharge_cms='surface_water__discharge',
              slope='topographic__slope'):
        """Erode into grid topography.

        For one time step, this erodes into the grid topography using
        the water discharge and topographic slope.

        The grid field 'topographic__elevation' is altered each time step.

        Parameters
        ----------
        dt : float
            Time step.
        discharge_cms : str, optional
            Name of the field that represents discharge on the nodes, if
            from the de Almeida solution have units of cubic meters per second.
        slope : str, optional
            Name of the field that represent topographic slope on each node.
        """
        try:
            S = self._grid.at_node[slope]
        except FieldError:
            raise ValueError('missing field for slope')

        if type(discharge_cms) is str:
            Q = self._grid.at_node[discharge_cms]
        else:
            Q = discharge_cms

        Q_to_m = np.power(Q, self.m)

        S_to_n = np.power(S, self.n)

        self.I = (self.K * Q_to_m * S_to_n) - self.entrainment_threshold

        self.I[self.I < 0.0] = 0.0

        self.dz = (self.uplift_rate - self.I) * dt

        self._grid['node'][elevs] += self.dz





# -*- coding: utf-8 -*-
"""
hex_lattice_tectonicizer.py.

Models discrete normal-fault offset on a 2D hex lattice with a rectangular
shape and with one orientation of the nodes being vertical.

The intention here is to use a particle (LCA) model to represent the evolution
of a 2D hillslope, with the hex_lattice_tectonicizer serving to shift the nodes
either upward (simple vertical uplift relative to baselevel), or up and
sideways (representing motion on a fault plane).

Created on Mon Nov 17 08:01:49 2014

@author: gtucker
"""

from landlab import HexModelGrid, ACTIVE_LINK, CORE_NODE
from landlab.core.utils import as_id_array
from numpy import (amax, zeros, arange, array, sqrt, where, logical_and,
                   logical_or, logical_xor, tan, cos, pi)
import numpy as np
from ..cfuncs import get_next_event_new  #, update_link_state_new

_DEFAULT_NUM_ROWS = 5
_DEFAULT_NUM_COLS = 5
_TAN60 = 1.732
_NEVER = 1.0e50  # this arbitrarily large val is also defined in ..cfuncs.pyx


def is_interior_link(link, grid):
    """Return True if both nodes are core; False otherwise."""
    return (grid.status_at_node[grid.node_at_link_tail[link]] == CORE_NODE
            and grid.status_at_node[grid.node_at_link_head[link]] == CORE_NODE)


def is_perim_link(link, grid):
    """Return True if both nodes are boundaries; False otherwise.

    Examples
    --------
    >>> from landlab import HexModelGrid
    >>> import numpy as np
    >>> mg = HexModelGrid(3, 4, 1.0, orientation='vertical', shape='rect')
    >>> is_perim_link(6, mg)
    True
    >>> is_perim_link(17, mg)
    False
    """
    return (grid.status_at_node[grid.node_at_link_tail[link]] != CORE_NODE
            and grid.status_at_node[grid.node_at_link_head[link]] != CORE_NODE)


class HexLatticeTectonicizer(object):
    """Handles tectonics and baselevel for CellLab-CTS models.

    This is the base class from which classes to represent particular
    baselevel/fault geometries are derived.

    Examples
    --------
    >>> hlt = HexLatticeTectonicizer()
    >>> hlt.grid.number_of_nodes
    25
    >>> hlt.nr
    5
    >>> hlt.nc
    5
    """

    def __init__(self, grid=None, node_state=None, propid=None, prop_data=None,
                 prop_reset_value=None):
        """
        Create and initialize a HexLatticeTectonicizer.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> hg = HexModelGrid(6, 6, shape='rect')
        >>> hlt = HexLatticeTectonicizer()
        >>> hlt.grid.number_of_nodes
        25
        >>> hlt.nr
        5
        >>> hlt.nc
        5
        """
        # If needed, create grid
        if grid is None:
            num_rows = _DEFAULT_NUM_ROWS
            num_cols = _DEFAULT_NUM_COLS
            self.grid = HexModelGrid(num_rows, num_cols, dx=1.0,
                                     orientation='vertical',
                                     shape='rect', reorient_links=True)
        else:
            # Make sure caller passed the right type of grid
            assert (grid.orientation == 'vertical'), \
                   'Grid must have vertical orientation'

            # Keep a reference to the grid
            self.grid = grid

        # If needed, create node-state grid
        if node_state is None:
            self.node_state = self.grid.add_zeros('node', 'node_state_map',
                                                  dtype=int)
        else:
            self.node_state = node_state

        # Remember the # of rows and cols
        self.nr = self.grid.number_of_node_rows
        self.nc = self.grid.number_of_node_columns

        # propid should be either a reference to a CA model's "property id"
        # array, or None
        self.propid = propid
        self.prop_data = prop_data
        self.prop_reset_value = prop_reset_value


class LatticeNormalFault(HexLatticeTectonicizer):
    """Handles normal-fault displacement in CellLab-CTS models.

    Represents a 60 degree, left-dipping normal fault, and handles discrete
    offsets for a hex grid that has vertical columns and a rectangular shape.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import HexModelGrid
    >>> from landlab.ca.boundaries.hex_lattice_tectonicizer import LatticeNormalFault
    >>> pid = np.arange(25, dtype=int)
    >>> pdata = np.arange(25)
    >>> ns = np.arange(25, dtype=int)
    >>> grid = HexModelGrid(5, 5, 1.0, orientation='vertical', shape='rect',
    ...                     reorient_links=True)
    >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
    >>> lnf.first_fw_col
    1
    >>> lnf.num_fw_rows
    array([0, 1, 3, 4, 5])
    >>> lnf.incoming_node
    array([1, 3, 4, 6])
    >>> lnf.outgoing_node
    array([12, 17, 19, 22])

    >>> pid = np.arange(16, dtype=int)
    >>> ns = np.arange(16, dtype=int)
    >>> pdata = np.arange(16)
    >>> grid = HexModelGrid(4, 4, 1.0, orientation='vertical', shape='rect',
    ...                     reorient_links=True)
    >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
    >>> lnf.num_fw_rows
    array([0, 1, 3, 4])
    >>> lnf.incoming_node
    array([1, 2, 5])
    >>> lnf.outgoing_node
    array([ 7, 11, 15])
    >>> lnf.do_offset(rock_state=16)
    >>> ns
    array([ 0, 16, 16, 16,  4, 16,  6,  1,  8,  2, 10,  5, 12, 13, 14,  9])
    >>> lnf.propid
    array([ 0,  7, 11,  3,  4, 15,  6,  1,  8,  2, 10,  5, 12, 13, 14,  9])

    >>> pid = np.arange(20, dtype=int)
    >>> ns = np.arange(20, dtype=int)
    >>> pdata = np.arange(20)
    >>> grid = HexModelGrid(4, 5, 1.0, orientation='vertical', shape='rect',
    ...                     reorient_links=True)
    >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
    >>> lnf.incoming_node
    array([1, 3, 4, 6])
    >>> lnf.outgoing_node
    array([12, 14, 17, 19])
    >>> lnf.do_offset(rock_state=20)
    >>> ns
    array([ 0, 20, 20, 20, 20,  5, 20, 20,  8,  1, 10,  3,  4, 13,  6, 15, 16,
            9, 18, 11])
    >>> lnf.propid
    array([ 0, 12,  2, 14, 17,  5, 19,  7,  8,  1, 10,  3,  4, 13,  6, 15, 16,
            9, 18, 11])
    """

    def __init__(self, fault_x_intercept=0.0, grid=None, node_state=None,
                 propid=None, prop_data=None, prop_reset_value=None):
        """
        Create and initialize a LatticeNormalFault object.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import HexModelGrid
        >>> from landlab.ca.boundaries.hex_lattice_tectonicizer import LatticeNormalFault

        >>> pid = np.arange(25, dtype=int)
        >>> pdata = np.arange(25)
        >>> ns = np.arange(25, dtype=int)
        >>> grid = HexModelGrid(5, 5, 1.0, orientation='vertical',
        ...                     shape='rect', reorient_links=True)
        >>> lnf = LatticeNormalFault(-0.01, grid, ns, pid, pdata, 0.0)
        >>> lnf.first_fw_col
        0
        >>> lnf.num_fw_rows
        array([1, 2, 4, 5, 5])
        >>> lnf.incoming_node
        array([0, 1, 3, 4, 6])
        >>> lnf.outgoing_node
        array([12, 17, 19, 22, 24])

        >>> pid = np.arange(16, dtype=int)
        >>> pdata = np.arange(16)
        >>> ns = np.arange(16, dtype=int)
        >>> grid = HexModelGrid(4, 4, 1.0, orientation='vertical',
        ...                     shape='rect', reorient_links=True)
        >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
        >>> lnf.first_fw_col
        1
        >>> lnf.num_fw_rows
        array([0, 1, 3, 4])
        >>> lnf.incoming_node
        array([1, 2, 5])
        >>> lnf.outgoing_node
        array([ 7, 11, 15])

        >>> pid = np.arange(45, dtype=int)
        >>> pdata = np.arange(45)
        >>> ns = np.arange(45, dtype=int)
        >>> grid = HexModelGrid(5, 9, 1.0, orientation='vertical',
        ...                     shape='rect', reorient_links=True)
        >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
        >>> lnf.first_fw_col
        1
        >>> lnf.num_fw_rows
        array([0, 1, 3, 4, 5, 5, 5, 5, 5])
        >>> lnf.incoming_node
        array([ 1,  2,  3,  5,  6,  7,  8, 10, 11, 12])
        >>> lnf.outgoing_node
        array([22, 31, 33, 34, 35, 38, 39, 40, 43, 44])
        """

        # Do the base class init
        super(LatticeNormalFault, self).__init__(grid, node_state, propid,
                                                 prop_data, prop_reset_value)
        # Set up data structures:
        #   Make sure the footwall location is such that the fault actually
        #   cuts across the grid. This means the x intercept has to be, at
        #   the very least, no smaller than the biggest x-coordinate, and if
        #   there is an even number of columns, it must be smaller than that
        #   number minus 1/tangent 60 degrees (0.57735)
        assert fault_x_intercept < (amax(self.grid.node_x) - 0.57735), 'err'

        #   Figure out which nodes are and are not within the footwall
        in_footwall = (self.grid.node_y < _TAN60 * (self.grid.node_x -
                       fault_x_intercept))

        # Set up array of link offsets: when slip occurs, what link's data get
        # copied into the present link? Which links get cut by fault plane and
        # need to have their states reset?
        self._setup_link_offsets(in_footwall)
        self._setup_links_to_update_after_offset(in_footwall)

        # Helpful to have an array of node IDs for the bottom full row. Because
        # the nodes in the bottom row are staggered in a vertical, rectangular
        # hex grid, the IDs go: 0, M, 1, M+1, 2, M+2, ... etc., where M is half
        # the number of columns, rounded up (so, for example, 3 for a 5- or 6-
        # column grid, etc.)
        half_num_cols = (self.nc + 1) // 2
        bottom_row_node_id = (arange(self.nc) // 2 +
                              (arange(self.nc) % 2) * half_num_cols)

        # Also useful to remember the number of even-numbered columns
        self.n_even_cols = (self.nc + 1) // 2

        #   Find the first of the bottom-row nodes that lies in the footwall.
        # This loop exploits the fact that nodes are numbered in an order
        # sorted by x then y, and that the bottom row is staggered, with node
        # zero being "low", like: ,',', etc.
        self.first_fw_col = 0
        n = 0
        while not in_footwall[bottom_row_node_id[n]]:
            n += 1
            self.first_fw_col += 1
            assert n < self.nc, 'overflow in loop'

        #   Remember the number of footwall rows in each column
        self.num_fw_rows = zeros(self.nc, dtype=int)
        for c in range(self.nc):
            current_row = 0
            while (current_row < self.nr and
                   in_footwall[bottom_row_node_id[c] + self.nc*current_row]):
                self.num_fw_rows[c] += 1
                current_row += 1

        # If we're handling properties and property IDs, we need to do some
        # setup
        if self.propid is not None:

            # We want to remember the node IDs of two sets of nodes: those
            # whose contents will vanish off the right-hand side (and possibly
            # the top) of the grid with each offset step, and those that gain
            # new ("rock") contents with each such step.
            #
            # First, let's find the latter set, which we'll call
            # "incoming_node" because material is flowing into these nodes from
            # below. To do this, we'll go column-by-column, starting with
            # the first column that has nodes in the footwall, and going to
            # the next-to-last column. Even-numbered columns have *two*
            # incoming nodes at their base (if the footwall reaches that high),
            # whereas odd-numbered columns have only one. Note that we'll start
            # with a list (so we can append), and then convert to an array
            # that belongs to this class.
            incoming_node_list = []
            lower_right_node = (self.nc % 2) * (self.nc // 2) + \
                               ((self.nc + 1) % 2) * (self.nc - 1)
            for n in range(0, self.nc + (self.nc + 1) // 2):
                if in_footwall[n] and ((n - lower_right_node) % self.nc) != 0:
                    incoming_node_list.append(n)

            # Convert to a numpy array that belongs to the class (so we can
            # use it in the do_offset() method).
            self.incoming_node = array(incoming_node_list, dtype=int)

            # Next, find the IDs the "outgoing" nodes. There will always be
            # some of these along the right side of the grid. Depending on the
            # height of the grid and the position of the fault, there may or
            # may not also be some along the top.
            #
            # To find out which nodes will be exiting the grid, we use
            # geometry: simply find out which nodes are (1) within the footwall
            # and (2) the tectonic offset would take them beyond the grid
            # perimeter. We already have information about the first condition.
            # For the second, let's start by defining the grid edges. The
            # y coordinates of the top full row of nodes will either be NR - 1
            # (even-numbered columns) or NR - 1/2. So we'll take as a
            # reasonable boundary line NR - 1/4: any node that would move above
            # this point is definitely out of the grid. Note that the vertical
            # offset of nodes during each earthquake will be 1.5, so if we were
            # to offset a top-row node, it would move to y = NR + 1/2 or
            # y = NR + 1. Odd-numbered columns in the next-from-top row will
            # move to y = NR, which is out of bounds, so we want to flag these
            # too, and therefore need our cutoff below this. Even-numbered
            # columns in the next-from-top row will end up at y = NR - 1/2,
            # which is within the grid. So our cutoff must be between NR - 1/2
            # and NR. Hence the choise of NR - 1/4 as the effective "top" of
            # the grid.
            top_grid_edge = self.nr - 0.25

            # The right-hand edge of the grid is a simpler case, because our
            # grid is vertically oriented and there is no stagger on the right
            # and left sides. So we simply make it half a column beyond the
            # right-most column. Column width is sqrt(3), the last column is
            # NC - 1, so the right edge y coordinate is (NC - 1/2) x sqrt(3)/2
            right_grid_edge = (self.nc - 0.5) * (sqrt(3.0) / 2.0)

            # To save a repeated calculation in a loop, we'll find a threshold
            # x and y coordinate beyond which any node offset would take them
            # off the grid.
            x_threshold = right_grid_edge - (sqrt(3.0) / 2.0)
            y_threshold = top_grid_edge - 1.5

            # Actually it turns out there is a third criterion. One or two
            # nodes in the lower-right corner could be counted as both
            # "incoming" (they're at the bottom of the grid) AND outgoing
            # (they're on the right-hand side). We ignored these in setting up
            # incoming, so we should ignore them for outgoing too. This is
            # easy: any nodes on the right side (x > x_threshold) that are also
            # near the bottom (y < 1.25) should be ignored.

            # Now march through all nodes, placing those on the list that meet
            # our criteria. Yes, it's slow to do this as a Python loop, but we
            # only do it once.
            outgoing_node_list = []
            for n in range(self.grid.number_of_nodes):
                if (((self.grid.node_x[n] > x_threshold and
                      self.grid.node_y[n] > 1.25) or
                     self.grid.node_y[n] > y_threshold) and in_footwall[n]):
                    outgoing_node_list.append(n)

            # Finally, convert the outgoing node list to an array stored in
            # this object
            self.outgoing_node = array(outgoing_node_list, dtype=int)

    def _link_in_footwall(self, link, node_in_footwall):
        """Return True of both nodes are in footwall, False otherwise."""
        return (node_in_footwall[self.grid.node_at_link_tail[link]]
                and node_in_footwall[self.grid.node_at_link_head[link]])

    def _get_link_orientation(self, link):
        """Return link orientation code for given link."""
        assert self.grid.orientation[0] == 'v', 'assumes vertical orientation'
        head = self.grid.node_at_link_head[link]
        tail = self.grid.node_at_link_tail[link]
        dx = self.grid.x_of_node[head] - self.grid.x_of_node[tail]
        dy = self.grid.y_of_node[head] - self.grid.y_of_node[tail]
        if dy > dx:
            return 0  # vertical
        elif dy > 0:
            return 1  # right and up
        else:
            return 2  # right and down

    def _setup_link_offsets(self, node_in_footwall):
        """Set up array with link IDs for shifting link data up and right.

        Notes
        -----
        The array contains the ID of the link TO WHICH the contents get
        shifted upon fault slip.

        Examples
        --------
        >>> from landlab.ca.boundaries.hex_lattice_tectonicizer import LatticeNormalFault
        >>> from landlab import HexModelGrid
        >>> pid = np.arange(25, dtype=int)
        >>> pdata = np.arange(25)
        >>> ns = np.arange(25, dtype=int)
        >>> grid = HexModelGrid(5, 5, 1.0, orientation='vertical',
        ...                     shape='rect')
        >>> lnf = LatticeNormalFault(-0.01, grid, ns, pid, pdata, 0.0)
        >>> lnf.link_offset_id[16:24]
        array([37, 17, 18, 19, 40, 21, 22, 43])
        >>> lnf.first_link_shifted_from
        16
        >>> lnf.first_link_shifted_to
        37
        
        >>> pid = np.arange(36, dtype=int)
        >>> pdata = np.arange(36)
        >>> ns = np.arange(36, dtype=int)
        >>> grid = HexModelGrid(6, 6, 1.0, orientation='vertical',
        ...                     shape='rect')
        >>> lnf = LatticeNormalFault(-0.1, grid, ns, pid, pdata, 0.0)
        >>> lnf.first_link_shifted_from
        19
        >>> lnf.first_link_shifted_to
        44
        >>> lnf.link_offset_id[19:41]
        array([44, 45, 21, 22, 23, 48, 25, 26, 52, 53, 29, 30, 31, 57, 33, 34, 35,
               61, 37, 38, 39, 64])
        """
        self.link_offset_id = arange(self.grid.number_of_links, dtype=np.int)
        nc = self.grid.number_of_node_columns
        default_offset = 2 * nc + 2 * (nc - 1) + nc // 2
        self.first_link_shifted_from = 0
        self.first_link_shifted_to = 0

        for ln in range(self.grid.number_of_links - (default_offset + 1)):
            if (self._link_in_footwall(ln, node_in_footwall)
                and is_interior_link(ln, self.grid)):
                tail_node = self.grid.node_at_link_tail[ln]
                (_, c) = self.grid.node_row_and_column(tail_node)
                link_orientation = self._get_link_orientation(ln)
                offset = default_offset
                if nc % 2 == 1:  # odd number of columns
                    if (link_orientation + ((c - 1) % 2)) == 2:
                        offset += 1
                else:  # even number of columns
                    if (c % 2) == 0 and link_orientation == 0:
                        offset -= 1
                if is_interior_link(ln + offset, self.grid):
                    self.link_offset_id[ln] = ln + offset
                    if self.first_link_shifted_from == 0:
                        self.first_link_shifted_from = ln
                        self.first_link_shifted_to = ln + offset

    def _setup_links_to_update_after_offset(self, in_footwall):
        """Create and store array with IDs of links for which to update
        transitions after fault offset.

        These are: all active boundary links with at least one node in the
        footwall, plus the lowest non-boundary links, including the
        next-to-lowest vertical links and those angling that are below them,
        plus the fault-crossing links.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> hg = HexModelGrid(5, 5, orientation='vert', shape='rect')
        >>> lu = LatticeNormalFault(fault_x_intercept=-0.01, grid=hg)
        >>> lu.first_link_shifted_to
        37
        >>> lu.links_to_update
        array([ 7, 10, 11, 13, 14, 15, 16, 17, 18, 20, 22, 23, 24, 25, 26, 27, 29,
               30, 31, 33, 36, 38, 42, 44, 46, 50, 51, 53])
        """
        g = self.grid
        lower_active = logical_and(arange(g.number_of_links) < self.first_link_shifted_to,
                                   g.status_at_link == ACTIVE_LINK)
        link_in_fw = logical_or(in_footwall[g.node_at_link_tail],
                                in_footwall[g.node_at_link_head])
        lower_active_fw = logical_and(lower_active, link_in_fw)
        active_bnd = logical_and(g.status_at_link == ACTIVE_LINK,
                        logical_or(g.status_at_node[g.node_at_link_tail] != 0,
                                   g.status_at_node[g.node_at_link_head] != 0))
        active_bnd_fw = logical_and(active_bnd, link_in_fw)
        crosses_fw = logical_and(g.status_at_link==ACTIVE_LINK,
                                 logical_xor(in_footwall[g.node_at_link_tail],
                                             in_footwall[g.node_at_link_head]))
        update = logical_or(logical_or(lower_active_fw, active_bnd_fw),
                            crosses_fw)
        self.links_to_update = as_id_array(where(update)[0])

    def assign_new_link_state_and_transition(self, link, ca, current_time):
        """Update state and schedule new transition for given link"""
        tail_state = ca.node_state[self.grid.node_at_link_tail[link]]
        head_state = ca.node_state[self.grid.node_at_link_head[link]]
        orientation = ca.link_orientation[link]
        new_link_state = int(orientation * ca.num_node_states_sq +
                             tail_state * ca.num_node_states + head_state)
        ca.update_link_state_new(link, new_link_state, current_time)

    def shift_scheduled_transitions(self, ca, current_time):
        """Update link IDs in scheduled events at offset links.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> from landlab.ca.oriented_hex_cts import OrientedHexCTS
        >>> from landlab.ca.celllab_cts import Transition
        >>> import numpy as np

        >>> mg = HexModelGrid(5, 5, 1.0, orientation='vertical', shape='rect')
        >>> nsd = {0 : 'yes', 1 : 'no'}
        >>> xnlist = []
        >>> xnlist.append(Transition((0,0,0), (1,1,0), 1.0, 'test'))
        >>> xnlist.append(Transition((0,0,1), (1,1,1), 1.0, 'test'))
        >>> xnlist.append(Transition((0,0,2), (1,1,2), 1.0, 'test'))
        >>> nsg = mg.add_zeros('node', 'node_state_grid')
        >>> pid = np.arange(25, dtype=int)
        >>> pdata = np.arange(25)
        >>> ohcts = OrientedHexCTS(mg, nsd, xnlist, nsg)
        >>> lnf = LatticeNormalFault(-0.1, grid=mg)
        >>> pq = ohcts.priority_queue._queue
        >>> (int(1000 * pq[11][0]), pq[11][1:])
        (752, (11, 23))
        >>> (int(1000 * pq[12][0]), pq[12][1:])
        (483, (9, 20))
        >>> (int(1000 * pq[30][0]), pq[30][1:])
        (575, (6, 16))
        >>> lnf.do_offset(ca=ohcts)
        >>> (int(1000 * pq[48][0]), pq[48][1:])
        (752, (11, 43))
        >>> (int(1000 * pq[54][0]), pq[54][1:])
        (483, (9, 40))
        >>> (int(1000 * pq[61][0]), pq[61][1:])
        (575, (6, 37))
        """
        for i in range(len(ca.priority_queue._queue)):
            link = ca.priority_queue._queue[i][2]
            if self.link_offset_id[link] != link:
                ca.priority_queue._queue[i] = (ca.priority_queue._queue[i][0],
                                               ca.priority_queue._queue[i][1],
                                               self.link_offset_id[link])

    def shift_link_states(self, ca, current_time):
        """Shift link data up and right.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> from landlab.ca.oriented_hex_cts import OrientedHexCTS
        >>> from landlab.ca.celllab_cts import Transition
        >>> import numpy as np

        >>> mg = HexModelGrid(5, 5, 1.0, orientation='vertical', shape='rect')
        >>> nsd = {0 : 'yes', 1 : 'no'}
        >>> xnlist = []
        >>> xnlist.append(Transition((1,0,0), (1,1,0), 1.0, 'frogging'))
        >>> xnlist.append(Transition((1,0,1), (1,1,1), 1.0, 'frogging'))
        >>> xnlist.append(Transition((1,0,2), (1,1,2), 1.0, 'frogging'))
        >>> nsg = mg.add_zeros('node', 'node_state_grid')
        >>> nsg[:10] = 1
        >>> pid = np.arange(25, dtype=int)
        >>> pdata = np.arange(25)
        >>> ohcts = OrientedHexCTS(mg, nsd, xnlist, nsg)
        >>> ohcts.link_state
        array([ 0,  0,  0,  0,  0,  0,  0,  3,  0,  0,  7, 11,  0,  3,  3,  7, 11,
                7, 11,  0,  2,  0,  9,  6,  9,  6,  2,  2,  4,  8,  4,  8,  0,  0,
                0,  8,  4,  8,  4,  0,  0,  4,  8,  4,  8,  0,  0,  0,  8,  4,  8,
                4,  0,  0,  0,  0,  0,  0,  0])
        >>> lnf = LatticeNormalFault(-0.01, mg, nsg, pid, pdata, 0.0)
        >>> lnf.first_link_shifted_from
        16
        >>> lnf.link_offset_id
        array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 37,
               17, 18, 19, 40, 21, 22, 43, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33,
               34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50,
               51, 52, 53, 54, 55, 56, 57, 58])
        >>> lnf.shift_link_states(ohcts, 0.0)
        >>> ohcts.link_state
        array([ 0,  0,  0,  0,  0,  0,  0,  3,  0,  0,  7, 11,  0,  3,  3,  7, 11,
                7, 11,  0,  2,  0,  9,  6,  9,  6,  2,  2,  4,  8,  4,  8,  0,  0,
                0,  8,  4, 11,  4,  0,  2,  4,  8,  6,  8,  0,  0,  0,  8,  4,  8,
                4,  0,  0,  0,  0,  0,  0,  0])
        """
        num_links = self.grid.number_of_links
        for lnk in range(num_links - 1, self.first_link_shifted_from - 1, -1):
            link_offset = self.link_offset_id[lnk]
            if link_offset != lnk:
                ca.link_state[link_offset] = ca.link_state[lnk]
                ca.next_trn_id[link_offset] = ca.next_trn_id[lnk]
                ca.next_update[link_offset] = ca.next_update[lnk]

        self.shift_scheduled_transitions(ca, current_time)

        for lnk in self.links_to_update:
            self.assign_new_link_state_and_transition(lnk, ca, current_time)

    def do_offset(self, ca=None, current_time=0.0, rock_state=1):
        """Apply 60-degree normal-fault offset.

        Offset is applied to a hexagonal grid with vertical node orientation
        and rectangular arrangement of nodes.

        Parameters
        ----------
        rock_state : int
            State code to apply to new cells introduced along bottom row.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.ca.boundaries.hex_lattice_tectonicizer import LatticeNormalFault
        >>> from landlab import HexModelGrid
        >>> from landlab.ca.oriented_hex_cts import OrientedHexCTS
        >>> from landlab.ca.celllab_cts import Transition

        >>> pid = np.arange(25, dtype=int)
        >>> pdata = np.arange(25)
        >>> ns = np.arange(25, dtype=int)
        >>> grid = HexModelGrid(5, 5, 1.0, orientation='vertical',
        ...                     shape='rect', reorient_links=True)
        >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
        >>> lnf.do_offset(rock_state=25)
        >>> ns
        array([ 0, 25, 25, 25, 25,  5, 25, 25,  8,  1, 10,  3,  4, 13,  6, 15, 16,
                9, 18, 11, 20, 21, 14, 23, 24])
        >>> lnf.propid
        array([ 0, 12,  2, 17, 19,  5, 22,  7,  8,  1, 10,  3,  4, 13,  6, 15, 16,
                9, 18, 11, 20, 21, 14, 23, 24])

        >>> ns[5:] = 0
        >>> ns[:5] = 1
        >>> nsd = {0 : 'yes', 1 : 'no'}
        >>> xnlist = []
        >>> xnlist.append(Transition((0,0,0), (1,1,0), 1.0, 'test'))
        >>> ohcts = OrientedHexCTS(grid, nsd, xnlist, ns)
        >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
        >>> ohcts.link_state[7:35]
        array([2, 0, 0, 6, 9, 0, 2, 2, 4, 8, 4, 8, 0, 0, 0, 8, 4, 8, 4, 0, 0, 4,
               8, 4, 8, 0, 0, 0])
        >>> lnf.do_offset(ca=ohcts, current_time=0.0, rock_state=1)
        >>> ohcts.link_state[7:35]
        array([ 3,  0,  0,  7, 11,  0,  2,  3,  4,  9,  7, 11,  0,  3,  0,  8,  5,
               11,  7,  0,  2,  4,  9,  6,  9,  0,  2,  0])
        """

        # If we need to shift the property ID numbers, we'll first need to
        # record the property IDs in those nodes that are about to "shift off
        # the grid" (or rather, their contents will shift) due to tectonic
        # motion. We'll call these "nodes to replace".
        if self.propid is not None:

            # Now, remember the property IDs of the nodes along the right side
            # and possibly top that are about to shift off the grid
            propids_for_incoming_nodes = self.propid[self.outgoing_node]

        # We go column-by-column, starting from the right side
        for c in range(self.grid.number_of_node_columns - 1,
                       self.first_fw_col - 1, -1):

            # Odd-numbered rows are shifted up in the hexagonal, vertically
            # oriented lattice
            row_offset = 2 - (c % 2)

            # Number of base nodes in the footwall in this column (1 or 2).
            n_base_nodes = min(self.num_fw_rows[c], row_offset)

            # ID of the bottom footwall node in this column
            bottom_node = (c // 2) + ((c % 2) * self.n_even_cols)

            # The bottom 1 or 2 nodes in this column are set to rock
            self.node_state[bottom_node] = rock_state
            if n_base_nodes == 2:
                self.node_state[bottom_node+self.nc] = rock_state

            # "indices" here contains the array indices of those nodes in this
            # column that are to be replaced by the ones in the column to the
            # left and down one or two nodes. We do this replacement if indices
            # contains any data.
            first_repl = bottom_node + n_base_nodes * self.nc
            last_repl = first_repl + (self.num_fw_rows[c] -
                                      (n_base_nodes + 1)) * self.nc
            indices = arange(first_repl, last_repl + 1, self.nc)

            if len(indices) > 0:
                offset = (self.nc + ((self.nc + 1) // 2) +
                          ((c + 1) % 2) * ((self.nc + 1) % 2))
                self.node_state[indices] = self.node_state[indices-offset]
                if self.propid is not None:
                    self.propid[indices] = self.propid[indices-offset]

        if self.propid is not None:
            self.propid[self.incoming_node] = propids_for_incoming_nodes
            self.prop_data[
                self.propid[self.incoming_node]] = self.prop_reset_value

        if ca is not None:
            self.shift_link_states(ca, current_time)



class LatticeUplifter(HexLatticeTectonicizer):
    """Handles vertical uplift of interior (not edges) for a hexagonal lattice
    with vertical node orientation and rectangular node arrangement.
    """
    def __init__(self, grid=None, node_state=None, propid=None, prop_data=None,
                 prop_reset_value=None, opt_block_layer=False, block_ID=9,
                 block_layer_dip_angle=0.0, block_layer_thickness=1.0,
                 layer_left_x=0.0, y0_top=0.0):
        """
        Create and initialize a LatticeUplifter

        Examples
        --------
        >>> lu = LatticeUplifter()
        >>> lu.inner_base_row_nodes
        array([1, 3, 4])

        >>> hg = HexModelGrid(5, 6, 1.0, orientation='vertical', shape='rect',
        ...                   reorient_links=True)
        >>> lu = LatticeUplifter(grid=hg)
        >>> lu.inner_base_row_nodes
        array([1, 2, 3, 4])
        """
        # Do the base class init
        super(LatticeUplifter, self).__init__(grid, node_state, propid,
                                              prop_data, prop_reset_value)

        # Remember the IDs of nodes on the bottom row
        self.inner_base_row_nodes = zeros(self.nc - 2, dtype=int)
        n_in_lower = (self.nc // 2) - 1  # num inner nodes bottom row
        upper_start = (self.nc + 1) // 2
        n_in_upper = upper_start - 1
        self.inner_base_row_nodes[:n_in_lower] = arange(1, n_in_lower + 1)
        self.inner_base_row_nodes[n_in_lower:] = arange(upper_start,
                                                        upper_start + \
                                                        n_in_upper)

        if self.propid is not None:
            self.inner_top_row_nodes = self.inner_base_row_nodes + \
                                       ((self.nr - 1) * self.nc)

        self._setup_links_to_update_after_uplift()

        # Handle option for a layer of "blocks"
        self.opt_block_layer = opt_block_layer
        if opt_block_layer:
            self.cum_uplift = 0.0
            self.block_ID = block_ID
            self.block_layer_thickness = block_layer_thickness
            self.block_layer_dip_angle = block_layer_dip_angle
            self.layer_left_x = layer_left_x
            self.y0_top = y0_top

    def _setup_links_to_update_after_uplift(self):
        """Create and store array with IDs of links for which to update
        transitions after uplift.

        These are: all active boundary links, plus the lowest non-boundary
        links, including the next-to-lowest vertical links and those angling
        that are below them.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> hg = HexModelGrid(6, 6, orientation='vert', shape='rect')
        >>> lu = LatticeUplifter(grid=hg)
        >>> lu.links_to_update
        array([ 8,  9, 11, 12, 13, 14, 15, 16, 18, 19, 20, 21, 22, 24, 25, 26, 30,
               34, 38, 42, 46, 50, 54, 58, 62, 66, 70, 72, 73, 74, 75, 76, 77, 79,
               80])
        >>> hg = HexModelGrid(5, 5, orientation='vert', shape='rect')
        >>> lu = LatticeUplifter(grid=hg)
        >>> lu.links_to_update
        array([ 7, 10, 11, 13, 14, 15, 16, 17, 18, 20, 22, 25, 28, 31, 35, 38, 41,
               44, 46, 48, 49, 50, 51, 52, 53])
        """
        g = self.grid
        nc = g.number_of_node_columns
        max_link_id = (3 * (nc - 1) + 2 * ((nc + 1) // 2) + nc // 2
                       + (nc - 1) // 2)
        lower_active = logical_and(arange(g.number_of_links) < max_link_id,
                                   g.status_at_link == 0)
        boundary = logical_or(g.status_at_node[g.node_at_link_tail] != 0,
                              g.status_at_node[g.node_at_link_head] != 0)
        active_bnd = logical_and(boundary, g.status_at_link == 0)
        self.links_to_update = as_id_array(where(logical_or(lower_active,
                                                            active_bnd))[0])

    def _get_new_base_nodes(self, rock_state):
        """
        Return an array (or scalar) of states for the newly uplifted bottom
        inner row.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> from landlab.ca.hex_cts import HexCTS
        >>> from landlab.ca.celllab_cts import Transition
        >>> mg = HexModelGrid(5, 5, 1.0, orientation='vertical', shape='rect')
        >>> nsd = {}  # node state dict
        >>> for i in range(10):
        ...     nsd[i] = i
        >>> xnlist = []
        >>> xnlist.append(Transition((0,0,0), (1,1,0), 1.0, 'frogging'))
        >>> nsg = mg.add_zeros('node', 'node_state_grid')
        >>> ca = HexCTS(mg, nsd, xnlist, nsg)

        >>> lu = LatticeUplifter(opt_block_layer=True)
        >>> lu._get_new_base_nodes(rock_state=7)
        array([9, 9, 9])
        >>> lu.uplift_interior_nodes(ca, current_time=0.0, rock_state=7)
        >>> lu.node_state[:5]
        array([0, 9, 0, 9, 9])
        >>> lu = LatticeUplifter(opt_block_layer=True, block_layer_thickness=2,
        ... block_layer_dip_angle=90.0, layer_left_x=1.0)
        >>> lu._get_new_base_nodes(rock_state=7)
        array([9, 7, 9])
        >>> lu.uplift_interior_nodes(ca, current_time=0.0, rock_state=7)
        >>> lu.node_state[:5]
        array([0, 9, 0, 7, 9])
        >>> lu = LatticeUplifter(opt_block_layer=True, block_layer_thickness=1,
        ... block_layer_dip_angle=45.0, y0_top=-1.0)
        >>> lu._get_new_base_nodes(rock_state=7)
        array([9, 7, 9])
        >>> lu.uplift_interior_nodes(ca, current_time=0.0, rock_state=7)
        >>> lu.node_state[:5]
        array([0, 9, 0, 7, 9])
        """

        new_base_nodes = zeros(len(self.inner_base_row_nodes), dtype=int)

        if self.block_layer_dip_angle == 0.0:  # horizontal

            if self.cum_uplift < self.block_layer_thickness:
                new_base_nodes[:] = self.block_ID
            else:
                new_base_nodes[:] = rock_state

        elif self.block_layer_dip_angle == 90.0:  # vertical

            layer_right_x = self.layer_left_x + self.block_layer_thickness
            inside_layer = where(logical_and(
                    self.grid.x_of_node[
                        self.inner_base_row_nodes] >= self.layer_left_x,
                    self.grid.x_of_node[
                        self.inner_base_row_nodes] <= layer_right_x))[0]
            new_base_nodes[:] = rock_state
            new_base_nodes[inside_layer] = self.block_ID

        else:

            x = self.grid.x_of_node[self.inner_base_row_nodes]
            y = self.grid.y_of_node[self.inner_base_row_nodes]
            m = tan(pi * self.block_layer_dip_angle / 180.0)
            y_top = m * x + self.y0_top
            y_bottom = y_top - (self.block_layer_thickness
                                / cos(pi * self.block_layer_dip_angle / 180.0))
            inside_layer = where(logical_and(y >= y_bottom, y <= y_top))
            new_base_nodes[:] = rock_state
            new_base_nodes[inside_layer] = self.block_ID

        return new_base_nodes

    def shift_link_and_transition_data_upward(self, ca, current_time):
        """Applies uplift to links and transitions.

        For each link that lies above the y = 1.5 cells line, assign the
        properties of the link one row down.
        
        (For an example, see unit test:
            test_shift_link_and_transition_data_upward)
        """

        # Find the ID of the first link above the y = 1.5 line
        nc = self.grid.number_of_node_columns
        first_link = (((nc - 1) // 2)      # skip bottom horizontals
                      + (3 * (nc - 1))     # skip 3 sets of diagonals
                      + nc                 # skip a full row of verticals
                      + ((nc + 1) // 2))   # skip a an even row of verticals

        # Define the offset in ID between a link and its neighbor one row up
        # (or down)
        shift = nc + 2 * (nc - 1)

        # Loop from top to bottom of grid, shifting the following link data
        # upward: state of link, ID of its next transition, and time of its
        # next transition.
        for lnk in range(self.grid.number_of_links - 1, first_link - 1, -1):
            ca.link_state[lnk] = ca.link_state[lnk - shift]
            ca.next_trn_id[lnk] = ca.next_trn_id[lnk - shift]
            ca.next_update[lnk] = ca.next_update[lnk - shift]

        # Sweep through event queue, shifting links upward. Do NOT shift links
        # with IDs greater than NL - [SHIFT + (NC - 1)], because these are so
        # close to the top of the grid that either the events would refer to
        # non-existent links (>= NL) or would involve shifting an event onto
        # an upper-boundary link. Note that because the event data are stored
        # in a tuple, we have to replace the entire tuple (can't simply change
        # the one item, because tuples are immutable)
        first_no_shift_id = self.grid.number_of_links - (shift + (nc - 1))
        for i in range(len(ca.priority_queue._queue)):
            if ca.priority_queue._queue[i][2] < first_no_shift_id:
                ca.priority_queue._queue[i] = (ca.priority_queue._queue[i][0],
                                               ca.priority_queue._queue[i][1],
                                               (ca.priority_queue._queue[i][2]
                                               + shift))

        # Update state of links along the boundaries.
        for lk in self.links_to_update:

            # Update link state
            fns = self.node_state[self.grid.node_at_link_tail[lk]]
            tns = self.node_state[self.grid.node_at_link_head[lk]]
            orientation = ca.link_orientation[lk]
            new_link_state = (orientation * ca.num_node_states_sq
                              + fns * ca.num_node_states + tns)

            # Schedule a new transition, if applicable
            ca.link_state[lk] = new_link_state
            if ca.n_trn[new_link_state] > 0:
                (event_time, this_trn_id) = get_next_event_new(lk,
                                                               new_link_state,
                                                               current_time,
                                                               ca.n_trn,
                                                               ca.trn_id,
                                                               ca.trn_rate)
                ca.priority_queue.push(lk, event_time)
                ca.next_update[lk] = event_time
                ca.next_trn_id[lk] = this_trn_id
            else:
                ca.next_update[lk] = _NEVER
                ca.next_trn_id[lk] = -1

    def uplift_property_ids(self):
        """
        Shift property IDs upward by one row
        """
        top_row_propid = self.propid[self.inner_top_row_nodes]
        for r in range(self.nr-1, 0, -1):
            self.propid[self.inner_base_row_nodes+self.nc*r] =  \
                    self.propid[self.inner_base_row_nodes+self.nc*(r-1)]
        self.propid[self.inner_base_row_nodes] = top_row_propid
        self.prop_data[
            self.propid[self.inner_base_row_nodes]] = self.prop_reset_value

    def uplift_interior_nodes(self, ca, current_time, rock_state=1):
        """
        Simulate 'vertical' displacement by shifting contents of node_state

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> from landlab.ca.hex_cts import HexCTS
        >>> from landlab.ca.celllab_cts import Transition
        >>> mg = HexModelGrid(5, 5, 1.0, orientation='vertical', shape='rect')
        >>> nsd = {}
        >>> for i in range(26):
        ...     nsd[i] = i
        >>> xnlist = []
        >>> xnlist.append(Transition((0,0,0), (1,1,0), 1.0, 'frogging', True))
        >>> nsg = mg.add_zeros('node', 'node_state_grid')
        >>> ca = HexCTS(mg, nsd, xnlist, nsg)
        >>> pd = mg.add_zeros('node', 'propdata')
        >>> lu = LatticeUplifter(propid=ca.propid, prop_data=pd)
        >>> lu.node_state[:] = np.arange(len(lu.node_state))
        >>> lu.uplift_interior_nodes(ca, rock_state=25, current_time=0.0)
        >>> lu.node_state # doctest: +NORMALIZE_WHITESPACE
        array([ 0, 25,  2, 25, 25,
                5,  1,  7,  3,  4,
               10,  6, 12,  8,  9,
               15, 11, 17, 13, 14,
               20, 16, 22, 18, 19])
        >>> lu.propid # doctest: +NORMALIZE_WHITESPACE
        array([ 0, 21,  2, 23, 24,
                5,  1,  7,  3,  4,
               10,  6, 12,  8,  9,
               15, 11, 17, 13, 14,
               20, 16, 22, 18, 19])
        """

        # Shift the node states up by a full row. A "full row" includes two
        # staggered rows.
        for r in range(self.nr - 1, 0, -1):
            # This row gets the contents of the nodes 1 row down
            self.node_state[self.inner_base_row_nodes+self.nc*r] = \
                    self.node_state[self.inner_base_row_nodes+self.nc*(r-1)]

        # Fill the bottom rows with "fresh material" (code = rock_state), or
        # if using a block layer, with the right pattern of states.
        if self.opt_block_layer:
            new_base_nodes = self._get_new_base_nodes(rock_state)
            self.cum_uplift += 1.0
            self.y0_top += 1.0
        else:
            new_base_nodes = rock_state
        self.node_state[self.inner_base_row_nodes] = new_base_nodes

        # If propid (property ID) is defined, shift that too.
        if self.propid is not None:
            self.uplift_property_ids()

        self.shift_link_and_transition_data_upward(ca, current_time)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
Created on Thu Jun 30 12:40:39 2016

@author: gtucker
"""

import numpy as np
cimport numpy as np
cimport cython
from landlab import CORE_NODE
from _heapq import heappush, heappop
from libc.stdlib cimport rand
from libc.math cimport log

import sys # for debug

cdef double _NEVER = 1.0e50

cdef int _CORE = CORE_NODE

DTYPE = np.double
ctypedef np.double_t DTYPE_t

DTYPE_INT = np.int
ctypedef np.int_t DTYPE_INT_t

DTYPE_INT8 = np.int8
ctypedef np.int8_t DTYPE_INT8_t

DTYPE_UINT8 = np.uint8
ctypedef np.uint8_t DTYPE_UINT8_t

cdef char _DEBUG = 0


cdef class PriorityQueue:
    """
    Implements a priority queue.
    """
    cdef public object _queue
    cdef public int _index

    def __init__(self):
        self._queue = []
        self._index = 0

    def push(self, int item, double priority):
        heappush(self._queue, (priority, self._index, item))
        self._index += 1

    def pop(self):
        assert len(self._queue) > 0, 'Q is empty'
        return heappop(self._queue)


cdef class Event:
    """
    Represents a transition event at a link. The transition occurs at a given
    link and a given time, and it involves a transition into the state xn_to
    (an integer code representing the new link state; "xn" is shorthand for
    "transition").

    The class overrides the __lt__ (less than operator) method so that when
    Event() objects are placed in a PriorityQueue, the earliest event is
    given the highest priority (i.e., placed at the top of the queue).

    Event() constructor sets 3 required properties and one optional
    property.

    Parameters
    ----------
    time : float
        Time at which the event is scheduled to occur
    link : int
        ID of the link at which event occurs
    xn_to : int
        New state to which this cell pair (link) will transition
    propswap : bool (optional)
        Flag: does this event involve an exchange of properties between
        the two cells?

    Examples
    --------
    >>> from landlab.ca.celllab_cts import Event
    >>> e1 = Event( 10.0, 1, 2)
    >>> e2 = Event( 2.0, 3, 1)
    >>> e1 < e2
    False
    >>> e2 < e1
    True
    """
    cdef public double time
    cdef public int link
    cdef public int xn_to
    cdef public char propswap
    cdef public object prop_update_fn

    def __init__(self, double time, int link, int xn_to, object propswap=False,
                 object prop_update_fn=None):
        """
        Event() constructor sets 3 required properties and one optional
        property.

        Parameters
        ----------
        time : float
            Time at which the event is scheduled to occur
        link : int
            ID of the link at which event occurs
        xn_to : int
            New state to which this cell pair (link) will transition
        propswap : bool (optional)
            Flag: does this event involve an exchange of properties between
            the two cells?
        """
        self.time = time
        self.link = link
        self.xn_to = xn_to
        self.propswap = propswap
        self.prop_update_fn = prop_update_fn
        
    def __richcmp__(self, Event other, int op):
        """
        Overridden less-than operator: returns true if the event on the left
        has an earlier scheduled time than the event on the right
        """
        return self.time < other.time


@cython.boundscheck(True)
@cython.wraparound(False)
cdef int current_link_state(DTYPE_INT_t link_id,
                       np.ndarray[DTYPE_INT_t, ndim=1] node_state, 
                       np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_tail,
                       np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_head,
                       np.ndarray[DTYPE_INT8_t, ndim=1] link_orientation,
                       DTYPE_INT_t num_node_states,
                       DTYPE_INT_t num_node_states_sq):
    """Get the current state of a link.

    Used to determine whether the link state at link *link_id* has changed
    due to an independent change in the node-state grid. Returns the
    current state of the link based on the states of its two end nodes;
    this can be compared to the entry in self.link_state to determine
    whether the state has changed.

    Parameters
    ----------
    link_id : int
        ID of the active link to test
    node_state : array of int
        State codes of nodes
    node_at_link_tail : array of int
        ID of node at link tails
    node_at_link_head : array of int
        ID of node at link heads
    link_orientation : array of 1-byte int
        Orientation codes: 0, 1, or (with hex) 2
    num_nodes_states : int
        Total number of possible node states
    num_node_states_sq : int
        Square of number of node states (precomputed for speed)
    
    Returns
    -------
    int
        New link state code
    """
    cdef int tail_node_state, head_node_state
    cdef char orientation

    # Find out the states of the two nodes, and the orientation
    tail_node_state = node_state[node_at_link_tail[link_id]]
    head_node_state = node_state[node_at_link_head[link_id]]
    orientation = link_orientation[link_id]

    # Return the corresponding state code.
    return (orientation * num_node_states_sq +
            tail_node_state * num_node_states + head_node_state)


@cython.boundscheck(True)
@cython.wraparound(False)
cpdef update_link_states_and_transitions(
                             np.ndarray[DTYPE_INT_t, ndim=1] active_links,
                             np.ndarray[DTYPE_INT_t, ndim=1] node_state, 
                             np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_tail,
                             np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_head,
                             np.ndarray[DTYPE_INT8_t, ndim=1] link_orientation,
                             np.ndarray[DTYPE_INT8_t, ndim=1] bnd_lnk,
                             np.ndarray[DTYPE_INT_t, ndim=1] link_state,
                             np.ndarray[DTYPE_INT_t, ndim=1] n_xn,
                             event_queue,
                             np.ndarray[DTYPE_t, ndim=1] next_update,
                             np.ndarray[DTYPE_INT_t, ndim=2] xn_to,
                             np.ndarray[DTYPE_t, ndim=2] xn_rate, 
                             DTYPE_INT_t num_node_states,
                             DTYPE_INT_t num_node_states_sq,
                             DTYPE_t current_time,
                             np.ndarray[DTYPE_INT8_t, ndim=2] xn_propswap,
                             xn_prop_update_fn):
        """
        Following an "external" change to the node state grid, updates link
        states where necessary and creates any needed events.

        Notes
        -----
        **Algorithm**::

            FOR each active link:
                if the actual node pair is different from the link's code:
                    change the link state to be correct
                    schedule an event
        """
        cdef int current_state
        cdef int i, j

        for j in range(len(active_links)):
            i = active_links[j]
            current_state = current_link_state(i, node_state,
                                               node_at_link_tail,
                                               node_at_link_head,
                                               link_orientation,
                                               num_node_states,
                                               num_node_states_sq)
            if current_state != link_state[i]:
                update_link_state(i, current_state, current_time, bnd_lnk,
                                  node_state, node_at_link_tail,
                                  node_at_link_head, link_orientation, 
                                  num_node_states, num_node_states_sq,
                                  link_state, n_xn, event_queue, next_update, 
                                  xn_to, xn_rate,xn_propswap,
                                  xn_prop_update_fn)


@cython.boundscheck(True)
@cython.wraparound(False)
cpdef update_link_states_and_transitions_new(
                             np.ndarray[DTYPE_INT_t, ndim=1] active_links,
                             np.ndarray[DTYPE_INT_t, ndim=1] node_state, 
                             np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_tail,
                             np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_head,
                             np.ndarray[DTYPE_INT8_t, ndim=1] link_orientation,
                             np.ndarray[DTYPE_INT8_t, ndim=1] bnd_lnk,
                             np.ndarray[DTYPE_INT_t, ndim=1] link_state,
                             np.ndarray[DTYPE_INT_t, ndim=1] n_trn,
                             priority_queue,
                             np.ndarray[DTYPE_t, ndim=1] next_update,
                             np.ndarray[DTYPE_INT_t, ndim=1] next_trn_id,
                             np.ndarray[DTYPE_INT_t, ndim=2] trn_id,
                             np.ndarray[DTYPE_t, ndim=1] trn_rate, 
                             DTYPE_INT_t num_node_states,
                             DTYPE_INT_t num_node_states_sq,
                             DTYPE_t current_time):
        """
        Following an "external" change to the node state grid, updates link
        states where necessary and creates any needed events.

        Notes
        -----
        **Algorithm**::

            FOR each active link:
                if the actual node pair is different from the link's code:
                    change the link state to be correct
                    schedule an event
        """
        cdef int current_state
        cdef int i, j

        for j in range(len(active_links)):
            i = active_links[j]
            current_state = current_link_state(i, node_state,
                                               node_at_link_tail,
                                               node_at_link_head,
                                               link_orientation,
                                               num_node_states,
                                               num_node_states_sq)
            if current_state != link_state[i]:
                update_link_state_new(i, current_state, current_time, bnd_lnk,
                                  node_state, node_at_link_tail,
                                  node_at_link_head, link_orientation, 
                                  num_node_states, num_node_states_sq,
                                  link_state, n_trn, priority_queue, next_update, 
                                  next_trn_id, trn_id, trn_rate)


@cython.boundscheck(True)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef update_node_states(np.ndarray[DTYPE_INT_t, ndim=1] node_state,
                       np.ndarray[DTYPE_UINT8_t, ndim=1] status_at_node,
                       DTYPE_INT_t tail_node, 
                       DTYPE_INT_t head_node,
                       DTYPE_INT_t new_link_state,
                       DTYPE_INT_t num_states):
    """Update the states of 2 nodes that underwent a transition."""
    if _DEBUG:
        print(('UNS', tail_node, head_node, new_link_state, num_states))
    # Change to the new states
    if status_at_node[tail_node] == _CORE:
        node_state[tail_node] = (new_link_state / num_states) % num_states # assume integer division!!
    if status_at_node[head_node] == _CORE:
        node_state[head_node] = new_link_state % num_states
    if _DEBUG:
        print(('UNS new tail state: ', node_state[tail_node]))
        print(('UNS new head state: ', node_state[head_node]))


@cython.boundscheck(True)
@cython.wraparound(False)
cpdef get_next_event(DTYPE_INT_t link, DTYPE_INT_t current_state, 
                   DTYPE_t current_time, 
                   np.ndarray[DTYPE_INT_t, ndim=1] n_xn,
                   np.ndarray[DTYPE_INT_t, ndim=2] xn_to,
                   np.ndarray[DTYPE_t, ndim=2] xn_rate,
                   np.ndarray[DTYPE_INT8_t, ndim=2] xn_propswap,
                   xn_prop_update_fn):
    """Get the next event for a link.

    Returns the next event for link with ID "link", which is in state
    "current state".

    Parameters
    ----------
    link : int
        ID of the link
    current_state : int
        Current state code for the link
    current_time : float
        Current time in simulation (i.e., time of event just processed)
    (see celllab_cts.py for other parameters)

    Returns
    -------
    Event object
        The returned Event object contains the time, link ID, and type of
        the next transition event at this link.

    Notes
    -----
    If there is only one potential transition out of the current state, a
    time for the transition is selected at random from an exponential
    distribution with rate parameter appropriate for this transition.

    If there are more than one potential transitions, a transition time is
    chosen for each, and the smallest of these applied.

    Assumes that there is at least one potential transition from the
    current state.
    """
    cdef int my_xn_to
    cdef int i
    cdef char propswap
    cdef double next_time, this_next
    cdef Event my_event

    # Find next event time for each potential transition
    if n_xn[current_state] == 1:
        my_xn_to = xn_to[current_state, 0]
        propswap = xn_propswap[current_state, 0]
        next_time = np.random.exponential(1.0 / xn_rate[current_state, 0])
        #next_time = -(1.0 / xn_rate[current_state, 0]) * log(1.0 - rand())
        prop_update_fn = xn_prop_update_fn[current_state, 0]
    else:
        next_time = _NEVER
        my_xn_to = 0
        propswap = 0
        for i in range(n_xn[current_state]):
            this_next = np.random.exponential(1.0 / xn_rate[current_state, i])
            #this_next = -(1.0 / xn_rate[current_state, i]) * log(1.0 - rand())
            if this_next < next_time:
                next_time = this_next
                my_xn_to = xn_to[current_state, i]
                propswap = xn_propswap[current_state, i]
                prop_update_fn = xn_prop_update_fn[current_state, i]

    # Create and setup event, and return it
    my_event = Event(next_time + current_time, link,
                     my_xn_to, propswap, prop_update_fn)

    return my_event


@cython.boundscheck(True)
@cython.wraparound(False)
cpdef get_next_event_new(DTYPE_INT_t link, DTYPE_INT_t current_state, 
                   DTYPE_t current_time, 
                   np.ndarray[DTYPE_INT_t, ndim=1] n_trn,
                   np.ndarray[DTYPE_INT_t, ndim=2] trn_id,
                   np.ndarray[DTYPE_t, ndim=1] trn_rate):
    """Get the next event for a link.

    Returns the next event for link with ID "link", which is in state
    "current state".

    Parameters
    ----------
    link : int
        ID of the link
    current_state : int
        Current state code for the link
    current_time : float
        Current time in simulation (i.e., time of event just processed)
    (see celllab_cts.py for other parameters)

    Returns
    -------
    Event object
        The returned Event object contains the time, link ID, and type of
        the next transition event at this link.

    Notes
    -----
    If there is only one potential transition out of the current state, a
    time for the transition is selected at random from an exponential
    distribution with rate parameter appropriate for this transition.

    If there are more than one potential transitions, a transition time is
    chosen for each, and the smallest of these applied.

    Assumes that there is at least one potential transition from the
    current state.
    """
    cdef int this_trn_id
    cdef int i
    cdef char propswap
    cdef double next_time, this_next

    # Find next event time for each potential transition
    if n_trn[current_state] == 1:
        this_trn_id = trn_id[current_state, 0]
        next_time = np.random.exponential(1.0 / trn_rate[this_trn_id])
    else:
        next_time = _NEVER
        this_trn_id = -1
        for i in range(n_trn[current_state]):
            this_next = np.random.exponential(1.0 / trn_rate[trn_id[current_state][i]])
            if this_next < next_time:
                next_time = this_next
                this_trn_id = trn_id[current_state, i]

    return (next_time + current_time, this_trn_id)


cpdef push_transitions_to_event_queue_new(int number_of_active_links,
                                          np.ndarray[DTYPE_INT_t, ndim=1] active_links,
                                          np.ndarray[DTYPE_INT_t, ndim=1] n_trn,
                                          np.ndarray[DTYPE_INT_t, ndim=1] link_state,
                                          np.ndarray[DTYPE_INT_t, ndim=2] trn_id,
                                          np.ndarray[DTYPE_t, ndim=1] trn_rate,
                                          np.ndarray[DTYPE_t, ndim=1] next_update,
                                          np.ndarray[DTYPE_INT_t, ndim=1] next_trn_id,
                                          PriorityQueue priority_queue):
    """
    Initializes the event queue by creating transition events for each
    cell pair that has one or more potential transitions and pushing these
    onto the queue. Also records scheduled transition times in the
    self.next_update array.
    """
    for j in range(number_of_active_links):

        i = active_links[j]
        if n_trn[link_state[i]] > 0:
            (ev_time, this_trn_id) = get_next_event_new(i, link_state[i], 0.0,
                                                        n_trn, trn_id,
                                                        trn_rate)
            priority_queue.push(i, ev_time)
            next_update[i] = ev_time
            next_trn_id[i] = this_trn_id

        else:
            next_update[i] = _NEVER

@cython.boundscheck(True)
@cython.wraparound(False)
cdef void update_link_state(DTYPE_INT_t link, DTYPE_INT_t new_link_state, 
                      DTYPE_t current_time,
                      np.ndarray[DTYPE_INT8_t, ndim=1] bnd_lnk,
                      np.ndarray[DTYPE_INT_t, ndim=1] node_state, 
                      np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_tail,
                      np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_head,
                      np.ndarray[DTYPE_INT8_t, ndim=1] link_orientation,
                      DTYPE_INT_t num_node_states,
                      DTYPE_INT_t num_node_states_sq,
                      np.ndarray[DTYPE_INT_t, ndim=1] link_state,
                      np.ndarray[DTYPE_INT_t, ndim=1] n_xn, event_queue,
                      np.ndarray[DTYPE_t, ndim=1] next_update,
                      np.ndarray[DTYPE_INT_t, ndim=2] xn_to,
                      np.ndarray[DTYPE_t, ndim=2] xn_rate, 
                      np.ndarray[DTYPE_INT8_t, ndim=2] xn_propswap,
                      np.ndarray[object, ndim=2] xn_prop_update_fn):
    """
    Implements a link transition by updating the current state of the link
    and (if appropriate) choosing the next transition event and pushing it
    on to the event queue.

    Parameters
    ----------
    link : int
        ID of the link to update
    new_link_state : int
        Code for the new state
    current_time : float
        Current time in simulation
    (see celllab_cts.py for other parameters)
    """
    cdef int fns, tns
    cdef int orientation
    cdef Event event

    # If the link connects to a boundary, we might have a different state
    # than the one we planned
    if bnd_lnk[link]:
        fns = node_state[node_at_link_tail[link]]
        tns = node_state[node_at_link_head[link]]
        orientation = link_orientation[link]
        new_link_state = orientation * num_node_states_sq + \
            fns * num_node_states + tns

    link_state[link] = new_link_state
    if n_xn[new_link_state] > 0:
        event = get_next_event(link, new_link_state, current_time, n_xn, xn_to,
                               xn_rate, xn_propswap, xn_prop_update_fn)
        heappush(event_queue, event)
        next_update[link] = event.time
    else:
        next_update[link] = _NEVER


@cython.boundscheck(True)
@cython.wraparound(False)
cdef void update_link_state_new(DTYPE_INT_t link, DTYPE_INT_t new_link_state, 
                      DTYPE_t current_time,
                      np.ndarray[DTYPE_INT8_t, ndim=1] bnd_lnk,
                      np.ndarray[DTYPE_INT_t, ndim=1] node_state, 
                      np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_tail,
                      np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_head,
                      np.ndarray[DTYPE_INT8_t, ndim=1] link_orientation,
                      DTYPE_INT_t num_node_states,
                      DTYPE_INT_t num_node_states_sq,
                      np.ndarray[DTYPE_INT_t, ndim=1] link_state,
                      np.ndarray[DTYPE_INT_t, ndim=1] n_trn,
                      PriorityQueue priority_queue,
                      np.ndarray[DTYPE_t, ndim=1] next_update,
                      np.ndarray[DTYPE_INT_t, ndim=1] next_trn_id,
                      np.ndarray[DTYPE_INT_t, ndim=2] trn_id,
                      np.ndarray[DTYPE_t, ndim=1] trn_rate):
    """
    Implements a link transition by updating the current state of the link
    and (if appropriate) choosing the next transition event and pushing it
    on to the event queue.

    Parameters
    ----------
    link : int
        ID of the link to update
    new_link_state : int
        Code for the new state
    current_time : float
        Current time in simulation
    (see celllab_cts.py for other parameters)
    """
    cdef int fns, tns
    cdef int this_trn_id
    cdef int orientation

    if _DEBUG:
        print(('ULSN', link, link_state[link], new_link_state, current_time))

    # If the link connects to a boundary, we might have a different state
    # than the one we planned
    if bnd_lnk[link]:
        fns = node_state[node_at_link_tail[link]]
        tns = node_state[node_at_link_head[link]]
        orientation = link_orientation[link]
        new_link_state = orientation * num_node_states_sq + \
            fns * num_node_states + tns
        if _DEBUG:
            print((' bnd True', new_link_state))

    link_state[link] = new_link_state
    if n_trn[new_link_state] > 0:
        (event_time, this_trn_id) = get_next_event_new(link, new_link_state, 
                                                       current_time,
                                                       n_trn, trn_id, trn_rate)
        priority_queue.push(link, event_time)
        next_update[link] = event_time
        next_trn_id[link] = this_trn_id
    else:
        next_update[link] = _NEVER
        next_trn_id[link] = -1

@cython.boundscheck(True)
@cython.wraparound(False)
cdef void do_transition(Event event,
                  np.ndarray[DTYPE_t, ndim=1] next_update,                  
                  np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_tail,                  
                  np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_head,                  
                  np.ndarray[DTYPE_INT_t, ndim=1] node_state,            
                  np.ndarray[DTYPE_INT_t, ndim=1] link_state,
                  np.ndarray[DTYPE_UINT8_t, ndim=1] status_at_node,
                  np.ndarray[DTYPE_INT8_t, ndim=1] link_orientation,
                  np.ndarray[DTYPE_INT_t, ndim=1] propid,
                  object prop_data,
                  np.ndarray[DTYPE_INT_t, ndim=1] n_xn,
                  np.ndarray[DTYPE_INT_t, ndim=2] xn_to,
                  np.ndarray[DTYPE_t, ndim=2] xn_rate, 
                  np.ndarray[DTYPE_INT_t, ndim=2] links_at_node,
                  np.ndarray[DTYPE_INT8_t, ndim=2] active_link_dirs_at_node,
                  DTYPE_INT_t num_node_states,
                  DTYPE_INT_t num_node_states_sq,
                  DTYPE_INT_t prop_reset_value,
                  np.ndarray[DTYPE_INT8_t, ndim=2] xn_propswap,
                  xn_prop_update_fn,
                  np.ndarray[DTYPE_INT8_t, ndim=1] bnd_lnk,
                  event_queue,
                  this_cts_model,
                  plot_each_transition=False,
                  plotter=None):
    """Transition state.

    Implements a state transition.

    Parameters
    ----------
    event : Event object
        Event object containing the data for the current transition event
    plot_each_transition : bool (optional)
        True if caller wants to show a plot of the grid after this
        transition
    plotter : CAPlotter object
        Sent if caller wants a plot after this transition
    (see celllab_cts.py for other parameters)

    Notes
    -----
    First checks that the transition is still valid by comparing the
    link's next_update time with the corresponding update time in the
    event object.

    If the transition is valid, we:

    1. Update the states of the two nodes attached to the link
    2. Update the link's state, choose its next transition, and push
       it on the event queue.
    3. Update the states of the other links attached to the two nodes,
       choose their next transitions, and push them on the event queue.
    """
    cdef int tail_node, head_node  # IDs of tail and head nodes at link
    cdef int old_tail_node_state
    cdef int old_head_node_state
    cdef int this_link_tail_node   # Tail ID for an adjacent link
    cdef int this_link_head_node   # Head ID for an adjacent link
    cdef int link                  # ID of a link
    cdef int new_link_state        # New link state after transition
    cdef int tmp                   # Used to exchange property IDs
    cdef char tail_changed, head_changed,  # Booleans
    cdef char dir_code             # Direction code for link at node
    cdef char orientation          # Orientation code for link
    cdef int i

    # We'll process the event if its update time matches the one we have
    # recorded for the link in question. If not, it means that the link has
    # changed state since the event was pushed onto the event queue, and
    # in that case we'll ignore it.
    if event.time == next_update[event.link]:

        tail_node = node_at_link_tail[event.link]
        head_node = node_at_link_head[event.link]

        # Remember the previous state of each node so we can detect whether the
        # state has changed
        old_tail_node_state = node_state[tail_node]
        old_head_node_state = node_state[head_node]

        update_node_states(node_state, status_at_node, tail_node,
                           head_node, event.xn_to, num_node_states)
        update_link_state(event.link, event.xn_to, event.time,
                          bnd_lnk, node_state,
                          node_at_link_tail,
                          node_at_link_head,
                          link_orientation, num_node_states,
                          num_node_states_sq, link_state,
                          n_xn, event_queue,
                          next_update, xn_to, xn_rate,
                          xn_propswap, xn_prop_update_fn)

        # Next, when the state of one of the link's nodes changes, we have
        # to update the states of the OTHER links attached to it. This
        # could happen to one or both nodes.
        if node_state[tail_node] != old_tail_node_state:

            for i in range(links_at_node.shape[1]):

                link = links_at_node[tail_node, i]
                dir_code = active_link_dirs_at_node[tail_node, i]

                if dir_code != 0 and link != event.link:

                    this_link_tail_node = node_at_link_tail[link]
                    this_link_head_node = node_at_link_head[link]
                    orientation = link_orientation[link]
                    new_link_state = (
                        orientation * num_node_states_sq +
                        node_state[this_link_tail_node] * num_node_states +
                        node_state[this_link_head_node])
                    update_link_state(link, new_link_state, event.time,
                                      bnd_lnk, node_state,
                                      node_at_link_tail,
                                      node_at_link_head,
                                      link_orientation, num_node_states,
                                      num_node_states_sq, link_state,
                                      n_xn, event_queue,
                                      next_update, xn_to, xn_rate,
                                      xn_propswap, xn_prop_update_fn)

        if node_state[head_node] != old_head_node_state:

            for i in range(links_at_node.shape[1]):

                link = links_at_node[head_node, i]
                dir_code = active_link_dirs_at_node[head_node, i]

                if dir_code != 0 and link != event.link:
                    this_link_tail_node = node_at_link_tail[link]
                    this_link_head_node = node_at_link_head[link]
                    orientation = link_orientation[link]
                    new_link_state = (
                        orientation * num_node_states_sq +
                        node_state[this_link_tail_node] * num_node_states +
                        node_state[this_link_head_node])
                    update_link_state(link, new_link_state, event.time,
                                      bnd_lnk, node_state,
                                      node_at_link_tail,
                                      node_at_link_head,
                                      link_orientation, num_node_states,
                                      num_node_states_sq, link_state,
                                      n_xn, event_queue,
                                      next_update, xn_to, xn_rate,
                                      xn_propswap, xn_prop_update_fn)

        # If requested, display a plot of the grid
        if plot_each_transition and (plotter is not None):
            plotter.update_plot()

        # If this event involves an exchange of properties (i.e., the
        # event involves motion of an object that posses properties we
        # want to track), implement the swap.
        #   If the event requires a call to a user-defined callback
        # function, we handle that here too.
        if event.propswap:
            tmp = propid[tail_node]
            propid[tail_node] = propid[head_node]
            propid[head_node] = tmp
            if status_at_node[tail_node] != _CORE:
                prop_data[propid[tail_node]] = prop_reset_value
            if status_at_node[head_node] != _CORE:
                prop_data[propid[head_node]] = prop_reset_value
            if event.prop_update_fn is not None:
                event.prop_update_fn(
                    this_cts_model, tail_node, head_node, event.time)

#@cython.boundscheck(False)
#@cython.wraparound(False)
cpdef void do_transition_new(DTYPE_INT_t event_link,
                  DTYPE_t event_time,
                  PriorityQueue priority_queue,
                  np.ndarray[DTYPE_t, ndim=1] next_update,                  
                  np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_tail,                  
                  np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_head,
                  np.ndarray[DTYPE_INT_t, ndim=1] node_state,
                  np.ndarray[DTYPE_INT_t, ndim=1] next_trn_id,
                  np.ndarray[DTYPE_INT_t, ndim=1] trn_to,
                  np.ndarray[DTYPE_UINT8_t, ndim=1] status_at_node,
                  DTYPE_INT_t num_node_states,
                  DTYPE_INT_t num_node_states_sq,
                  np.ndarray[DTYPE_INT8_t, ndim=1] bnd_lnk,
                  np.ndarray[DTYPE_INT8_t, ndim=1] link_orientation,
                  np.ndarray[DTYPE_INT_t, ndim=1] link_state,
                  np.ndarray[DTYPE_INT_t, ndim=1] n_trn,
                  np.ndarray[DTYPE_INT_t, ndim=2] trn_id,
                  np.ndarray[DTYPE_t, ndim=1] trn_rate, 
                  np.ndarray[DTYPE_INT_t, ndim=2] links_at_node,
                  np.ndarray[DTYPE_INT8_t, ndim=2] active_link_dirs_at_node,
                  np.ndarray[DTYPE_INT8_t, ndim=1] trn_propswap,
                  np.ndarray[DTYPE_INT_t, ndim=1] propid,
                  object prop_data,
                  DTYPE_INT_t prop_reset_value,
                  object trn_prop_update_fn,
                  object this_cts_model,
                  plot_each_transition=False,
                  plotter=None):
    """Transition state.

    Implements a state transition.

    Parameters
    ----------
    event : Event object
        Event object containing the data for the current transition event
    plot_each_transition : bool (optional)
        True if caller wants to show a plot of the grid after this
        transition
    plotter : CAPlotter object
        Sent if caller wants a plot after this transition
    (see celllab_cts.py for other parameters)

    Notes
    -----
    First checks that the transition is still valid by comparing the
    link's next_update time with the corresponding update time in the
    event object.

    If the transition is valid, we:

    1. Update the states of the two nodes attached to the link
    2. Update the link's state, choose its next transition, and push
       it on the event queue.
    3. Update the states of the other links attached to the two nodes,
       choose their next transitions, and push them on the event queue.
    """
    cdef int index
    cdef int tail_node, head_node  # IDs of tail and head nodes at link
    cdef int old_tail_node_state
    cdef int old_head_node_state
    cdef int this_trn_id
    cdef int this_trn_to
    cdef int this_link_tail_node   # Tail ID for an adjacent link
    cdef int this_link_head_node   # Head ID for an adjacent link
    cdef int link                  # ID of a link
    cdef int new_link_state        # New link state after transition
    cdef int tmp                   # Used to exchange property IDs
    cdef char tail_changed, head_changed,  # Booleans
    cdef char dir_code             # Direction code for link at node
    cdef char orientation          # Orientation code for link
    cdef int i

    if _DEBUG:
        print(('DTN', event_time, event_link, link_state[event_link], next_update[event_link]))

    # We'll process the event if its update time matches the one we have
    # recorded for the link in question. If not, it means that the link has
    # changed state since the event was pushed onto the event queue, and
    # in that case we'll ignore it.
    if event_time == next_update[event_link]:

        tail_node = node_at_link_tail[event_link]
        head_node = node_at_link_head[event_link]

        # DEBUG
        if status_at_node[tail_node] == 4 or status_at_node[head_node] == 4:
            print(('TRN INFO: ', event_time, event_link, link_state[event_link], next_update[event_link]))
            print('TAIL ' + str(tail_node) + ' ' + status_at_node[tail_node])
            print('HEAD ' + str(head_node) + ' ' + status_at_node[tail_node])
            #_DEBUG = True

        # Remember the previous state of each node so we can detect whether the
        # state has changed
        old_tail_node_state = node_state[tail_node]
        old_head_node_state = node_state[head_node]

        this_trn_id = next_trn_id[event_link]
        this_trn_to = trn_to[this_trn_id]

        if _DEBUG:
            print((this_trn_id, this_trn_to))
            print(('tail:', tail_node))
            print(('tail state:', old_tail_node_state))
            print(('head:', head_node))
            print(('head state:', old_head_node_state))

        update_node_states(node_state, status_at_node, tail_node,
                           head_node, this_trn_to, num_node_states)
        update_link_state_new(event_link, this_trn_to, event_time, bnd_lnk,
                              node_state, node_at_link_tail,
                              node_at_link_head, link_orientation,
                              num_node_states, num_node_states_sq,
                              link_state, n_trn, priority_queue, next_update,
                              next_trn_id, trn_id, trn_rate)

        # Next, when the state of one of the link's nodes changes, we have
        # to update the states of the OTHER links attached to it. This
        # could happen to one or both nodes.
        if node_state[tail_node] != old_tail_node_state:

            for i in range(links_at_node.shape[1]):

                link = links_at_node[tail_node, i]
                dir_code = active_link_dirs_at_node[tail_node, i]

                if dir_code != 0 and link != event_link:

                    this_link_tail_node = node_at_link_tail[link]
                    this_link_head_node = node_at_link_head[link]
                    orientation = link_orientation[link]
                    new_link_state = (
                        orientation * num_node_states_sq +
                        node_state[this_link_tail_node] * num_node_states +
                        node_state[this_link_head_node])
                    update_link_state_new(link, new_link_state, event_time, 
                                          bnd_lnk,
                                          node_state, node_at_link_tail,
                                          node_at_link_head, link_orientation,
                                          num_node_states, num_node_states_sq,
                                          link_state, n_trn, priority_queue, next_update,
                                          next_trn_id, trn_id, trn_rate)

        if node_state[head_node] != old_head_node_state:

            for i in range(links_at_node.shape[1]):

                link = links_at_node[head_node, i]
                dir_code = active_link_dirs_at_node[head_node, i]

                if dir_code != 0 and link != event_link:
                    this_link_tail_node = node_at_link_tail[link]
                    this_link_head_node = node_at_link_head[link]
                    orientation = link_orientation[link]
                    new_link_state = (
                        orientation * num_node_states_sq +
                        node_state[this_link_tail_node] * num_node_states +
                        node_state[this_link_head_node])
                    update_link_state_new(link, new_link_state, event_time, 
                                          bnd_lnk,
                                          node_state, node_at_link_tail,
                                          node_at_link_head, link_orientation,
                                          num_node_states, num_node_states_sq,
                                          link_state, n_trn, priority_queue, next_update,
                                          next_trn_id, trn_id, trn_rate)

        # If requested, display a plot of the grid
        if plot_each_transition and (plotter is not None):
            plotter.update_plot()

        # If this event involves an exchange of properties (i.e., the
        # event involves motion of an object that posses properties we
        # want to track), implement the swap.
        #   If the event requires a call to a user-defined callback
        # function, we handle that here too.
        if trn_propswap[this_trn_id]:
            tmp = propid[tail_node]
            propid[tail_node] = propid[head_node]
            propid[head_node] = tmp
            if status_at_node[tail_node] != _CORE:
                prop_data[propid[tail_node]] = prop_reset_value
            if status_at_node[head_node] != _CORE:
                prop_data[propid[head_node]] = prop_reset_value
            if trn_prop_update_fn[this_trn_id] != 0:
                trn_prop_update_fn[this_trn_id](
                    this_cts_model, tail_node, head_node, event_time)

cpdef double run_cts_new(double run_to, double current_time,
                     PriorityQueue priority_queue,
                     np.ndarray[DTYPE_t, ndim=1] next_update,                  
                     np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_tail,                  
                     np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_head,                  
                     np.ndarray[DTYPE_INT_t, ndim=1] node_state,            
                     np.ndarray[DTYPE_INT_t, ndim=1] next_trn_id,            
                     np.ndarray[DTYPE_INT_t, ndim=1] trn_to,
                     np.ndarray[DTYPE_UINT8_t, ndim=1] status_at_node,
                     DTYPE_INT_t num_node_states,
                     DTYPE_INT_t num_node_states_sq,
                     np.ndarray[DTYPE_INT8_t, ndim=1] bnd_lnk,
                     np.ndarray[DTYPE_INT8_t, ndim=1] link_orientation,
                     np.ndarray[DTYPE_INT_t, ndim=1] link_state,
                     np.ndarray[DTYPE_INT_t, ndim=1] n_trn,
                     np.ndarray[DTYPE_INT_t, ndim=2] trn_id,
                     np.ndarray[DTYPE_t, ndim=1] trn_rate, 
                     np.ndarray[DTYPE_INT_t, ndim=2] links_at_node,
                     np.ndarray[DTYPE_INT8_t, ndim=2] active_link_dirs_at_node,
                     np.ndarray[DTYPE_INT8_t, ndim=1] trn_propswap,
                     np.ndarray[DTYPE_INT_t, ndim=1] propid,
                     object prop_data,
                     DTYPE_INT_t prop_reset_value,
                     trn_prop_update_fn,
                     this_cts_model,
                     char plot_each_transition,
                     object plotter):
    """Run the model forward for a specified period of time.

    Parameters
    ----------
    run_to : float
        Time to run to, starting from self.current_time
    node_state_grid : 1D array of ints (x number of nodes) (optional)
        Node states (if given, replaces model's current node state grid)
    plot_each_transition : bool (optional)
        Option to display the grid after each transition
    plotter : CAPlotter object (optional)
        Needed if caller wants to plot after every transition
    (see celllab_cts.py for other parameters)
    """
    import sys
    cdef double ev_time
    cdef int ev_idx
    cdef int ev_link

    # Continue until we've run out of either time or events
    while current_time < run_to and priority_queue._queue:

        if _DEBUG:
            print('current time = ', current_time)

        # Is there an event scheduled to occur within this run?
        if priority_queue._queue[0][0] <= run_to:

            # If so, pick the next transition event from the event queue
            (ev_time, ev_idx, ev_link) = priority_queue.pop()

            if _DEBUG:
                print('event:', ev_time, ev_link, trn_to[next_trn_id[ev_link]])
                print('pq top is now:', priority_queue._queue[0])

            # ... and execute the transition
            do_transition_new(ev_link, ev_time, priority_queue, next_update,
                              node_at_link_tail,
                              node_at_link_head,
                              node_state, 
                              next_trn_id,
                              trn_to,
                              status_at_node,
                              num_node_states,
                              num_node_states_sq,
                              bnd_lnk,
                              link_orientation,
                              link_state,
                              n_trn,
                              trn_id,
                              trn_rate,
                              links_at_node,
                              active_link_dirs_at_node,
                              trn_propswap,
                              propid, prop_data,
                              prop_reset_value,
                              trn_prop_update_fn,
                              this_cts_model,
                              plot_each_transition,
                              plotter)

            # Update current time
            current_time = ev_time

        # If there is no event scheduled for this span of time, simply
        # advance current_time to the end of the current run period.
        else:
            current_time = run_to

        if _DEBUG:
            print(node_state)

    return current_time


cpdef double run_cts(double run_to, double current_time,
                     char plot_each_transition,
                     object plotter,
                     object event_queue,
                     np.ndarray[DTYPE_t, ndim=1] next_update,                  
                     np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_tail,                  
                     np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_head,                  
                     np.ndarray[DTYPE_INT_t, ndim=1] node_state,            
                     np.ndarray[DTYPE_INT_t, ndim=1] link_state,
                     np.ndarray[DTYPE_UINT8_t, ndim=1] status_at_node,
                     np.ndarray[DTYPE_INT8_t, ndim=1] link_orientation,
                     np.ndarray[DTYPE_INT_t, ndim=1] propid,
                     object prop_data,
                     np.ndarray[DTYPE_INT_t, ndim=1] n_xn,
                     np.ndarray[DTYPE_INT_t, ndim=2] xn_to,
                     np.ndarray[DTYPE_t, ndim=2] xn_rate, 
                     np.ndarray[DTYPE_INT_t, ndim=2] links_at_node,
                     np.ndarray[DTYPE_INT8_t, ndim=2] active_link_dirs_at_node,
                     DTYPE_INT_t num_node_states,
                     DTYPE_INT_t num_node_states_sq,
                     DTYPE_INT_t prop_reset_value,
                     np.ndarray[DTYPE_INT8_t, ndim=2] xn_propswap,
                     xn_prop_update_fn,
                     np.ndarray[DTYPE_INT8_t, ndim=1] bnd_lnk,
                     this_cts_model):
    """Run the model forward for a specified period of time.

    Parameters
    ----------
    run_to : float
        Time to run to, starting from self.current_time
    node_state_grid : 1D array of ints (x number of nodes) (optional)
        Node states (if given, replaces model's current node state grid)
    plot_each_transition : bool (optional)
        Option to display the grid after each transition
    plotter : CAPlotter object (optional)
        Needed if caller wants to plot after every transition
    (see celllab_cts.py for other parameters)
    """
    cdef Event ev

    # Continue until we've run out of either time or events
    while current_time < run_to and event_queue:

        # Is there an event scheduled to occur within this run?
        if event_queue[0].time <= run_to:

            # If so, pick the next transition event from the event queue
            ev = heappop(event_queue)

            # ... and execute the transition
            do_transition(ev, next_update,
                              node_at_link_tail,
                              node_at_link_head,
                              node_state, link_state,
                              status_at_node, link_orientation,
                              propid, prop_data,
                              n_xn, xn_to, xn_rate,
                              links_at_node,
                              active_link_dirs_at_node,
                              num_node_states, num_node_states_sq,
                              prop_reset_value, xn_propswap,
                              xn_prop_update_fn,
                              bnd_lnk, event_queue,
                              this_cts_model,
                              plot_each_transition,
                              plotter)

            # Update current time
            current_time = ev.time
            
        # If there is no event scheduled for this span of time, simply
        # advance current_time to the end of the current run period.
        else:
            current_time = run_to

    return current_time


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef get_next_event_lean(DTYPE_INT_t link, DTYPE_INT_t current_state, 
                   DTYPE_t current_time, 
                   np.ndarray[DTYPE_INT_t, ndim=1] n_xn,
                   np.ndarray[DTYPE_INT_t, ndim=2] xn_to,
                   np.ndarray[DTYPE_t, ndim=2] xn_rate):
    """Get the next event for a link.

    Returns the next event for link with ID "link", which is in state
    "current state". This "lean" version omits parameters related to property
    exchange and callback function.

    Parameters
    ----------
    link : int
        ID of the link
    current_state : int
        Current state code for the link
    current_time : float
        Current time in simulation (i.e., time of event just processed)
    (see celllab_cts.py for other parameters)

    Returns
    -------
    Event object
        The returned Event object contains the time, link ID, and type of
        the next transition event at this link.

    Notes
    -----
    If there is only one potential transition out of the current state, a
    time for the transition is selected at random from an exponential
    distribution with rate parameter appropriate for this transition.

    If there are more than one potential transitions, a transition time is
    chosen for each, and the smallest of these applied.

    Assumes that there is at least one potential transition from the
    current state.
    """
    cdef int my_xn_to
    cdef int i
    cdef char propswap
    cdef double next_time, this_next
    cdef Event my_event

    # Find next event time for each potential transition
    if n_xn[current_state] == 1:
        my_xn_to = xn_to[current_state, 0]
        next_time = np.random.exponential(1.0 / xn_rate[current_state, 0])
        #next_time = -(1.0 / xn_rate[current_state, 0]) * log(1.0 - rand())
    else:
        next_time = _NEVER
        my_xn_to = 0
        for i in range(n_xn[current_state]):
            this_next = np.random.exponential(1.0 / xn_rate[current_state, i])
            #this_next = -(1.0 / xn_rate[current_state, i]) * log(1.0 - rand())
            if this_next < next_time:
                next_time = this_next
                my_xn_to = xn_to[current_state, i]

    # Create and setup event, and return it
    my_event = Event(next_time + current_time, link, my_xn_to)

    return my_event


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void update_link_state_lean(DTYPE_INT_t link, DTYPE_INT_t new_link_state, 
                      DTYPE_t current_time,
                      np.ndarray[DTYPE_INT8_t, ndim=1] bnd_lnk,
                      np.ndarray[DTYPE_INT_t, ndim=1] node_state, 
                      np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_tail,
                      np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_head,
                      np.ndarray[DTYPE_INT8_t, ndim=1] link_orientation,
                      DTYPE_INT_t num_node_states,
                      DTYPE_INT_t num_node_states_sq,
                      np.ndarray[DTYPE_INT_t, ndim=1] link_state,
                      np.ndarray[DTYPE_INT_t, ndim=1] n_xn, event_queue,
                      np.ndarray[DTYPE_t, ndim=1] next_update,
                      np.ndarray[DTYPE_INT_t, ndim=2] xn_to,
                      np.ndarray[DTYPE_t, ndim=2] xn_rate):
    """
    Implements a link transition by updating the current state of the link
    and (if appropriate) choosing the next transition event and pushing it
    on to the event queue. This "lean" version omits parameters related to 
    property exchange and callback function.

    Parameters
    ----------
    link : int
        ID of the link to update
    new_link_state : int
        Code for the new state
    current_time : float
        Current time in simulation
    (see celllab_cts.py for other parameters)
    """
    cdef int fns, tns
    cdef int orientation
    cdef Event event

    # If the link connects to a boundary, we might have a different state
    # than the one we planned
    if bnd_lnk[link]:
        fns = node_state[node_at_link_tail[link]]
        tns = node_state[node_at_link_head[link]]
        orientation = link_orientation[link]
        new_link_state = orientation * num_node_states_sq + \
            fns * num_node_states + tns

    link_state[link] = new_link_state
    if n_xn[new_link_state] > 0:
        event = get_next_event_lean(link, new_link_state, current_time, n_xn, xn_to,
                               xn_rate)
        heappush(event_queue, event)
        next_update[link] = event.time
    else:
        next_update[link] = _NEVER


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void do_transition_lean(Event event,
                  np.ndarray[DTYPE_t, ndim=1] next_update,                  
                  np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_tail,                  
                  np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_head,                  
                  np.ndarray[DTYPE_INT_t, ndim=1] node_state,            
                  np.ndarray[DTYPE_INT_t, ndim=1] link_state,
                  np.ndarray[DTYPE_UINT8_t, ndim=1] status_at_node,
                  np.ndarray[DTYPE_INT8_t, ndim=1] link_orientation,
                  np.ndarray[DTYPE_INT_t, ndim=1] n_xn,
                  np.ndarray[DTYPE_INT_t, ndim=2] xn_to,
                  np.ndarray[DTYPE_t, ndim=2] xn_rate, 
                  np.ndarray[DTYPE_INT_t, ndim=2] links_at_node,
                  np.ndarray[DTYPE_INT8_t, ndim=2] active_link_dirs_at_node,
                  DTYPE_INT_t num_node_states,
                  DTYPE_INT_t num_node_states_sq,
                  np.ndarray[DTYPE_INT8_t, ndim=1] bnd_lnk,
                  object event_queue):
    """Transition state.

    Implements a state transition. This "lean" version omits parameters related
    to property exchange and callback function.

    Parameters
    ----------
    event : Event object
        Event object containing the data for the current transition event
    (see celllab_cts.py for other parameters)

    Notes
    -----
    First checks that the transition is still valid by comparing the
    link's next_update time with the corresponding update time in the
    event object.

    If the transition is valid, we:

    1. Update the states of the two nodes attached to the link
    2. Update the link's state, choose its next transition, and push
       it on the event queue.
    3. Update the states of the other links attached to the two nodes,
       choose their next transitions, and push them on the event queue.
    """
    cdef int tail_node, head_node  # IDs of tail and head nodes at link
    cdef int old_tail_node_state
    cdef int old_head_node_state
    cdef int this_link_tail_node   # Tail ID for an adjacent link
    cdef int this_link_head_node   # Head ID for an adjacent link
    cdef int link                  # ID of a link
    cdef int new_link_state        # New link state after transition
    cdef int tmp                   # Used to exchange property IDs
    cdef char tail_changed, head_changed,  # Booleans
    cdef char dir_code             # Direction code for link at node
    cdef char orientation          # Orientation code for link
    cdef int i

    #print 'dtl'

    # We'll process the event if its update time matches the one we have
    # recorded for the link in question. If not, it means that the link has
    # changed state since the event was pushed onto the event queue, and
    # in that case we'll ignore it.
    if event.time == next_update[event.link]:

        tail_node = node_at_link_tail[event.link]
        head_node = node_at_link_head[event.link]

        # Remember the previous state of each node so we can detect whether the
        # state has changed
        old_tail_node_state = node_state[tail_node]
        old_head_node_state = node_state[head_node]

        update_node_states(node_state, status_at_node, tail_node,
                           head_node, event.xn_to, num_node_states)
        update_link_state_lean(event.link, event.xn_to, event.time,
                          bnd_lnk, node_state,
                          node_at_link_tail,
                          node_at_link_head,
                          link_orientation, num_node_states,
                          num_node_states_sq, link_state,
                          n_xn, event_queue,
                          next_update, xn_to, xn_rate)

        # Next, when the state of one of the link's nodes changes, we have
        # to update the states of the OTHER links attached to it. This
        # could happen to one or both nodes.
        if node_state[tail_node] != old_tail_node_state:

            for i in range(links_at_node.shape[1]):

                link = links_at_node[tail_node, i]
                dir_code = active_link_dirs_at_node[tail_node, i]

                if dir_code != 0 and link != event.link:

                    this_link_tail_node = node_at_link_tail[link]
                    this_link_head_node = node_at_link_head[link]
                    orientation = link_orientation[link]
                    new_link_state = (
                        orientation * num_node_states_sq +
                        node_state[this_link_tail_node] * num_node_states +
                        node_state[this_link_head_node])
                    update_link_state_lean(link, new_link_state, event.time,
                                      bnd_lnk, node_state,
                                      node_at_link_tail,
                                      node_at_link_head,
                                      link_orientation, num_node_states,
                                      num_node_states_sq, link_state,
                                      n_xn, event_queue,
                                      next_update, xn_to, xn_rate)

        if node_state[head_node] != old_head_node_state:

            for i in range(links_at_node.shape[1]):

                link = links_at_node[head_node, i]
                dir_code = active_link_dirs_at_node[head_node, i]

                if dir_code != 0 and link != event.link:
                    this_link_tail_node = node_at_link_tail[link]
                    this_link_head_node = node_at_link_head[link]
                    orientation = link_orientation[link]
                    new_link_state = (
                        orientation * num_node_states_sq +
                        node_state[this_link_tail_node] * num_node_states +
                        node_state[this_link_head_node])
                    update_link_state_lean(link, new_link_state, event.time,
                                      bnd_lnk, node_state,
                                      node_at_link_tail,
                                      node_at_link_head,
                                      link_orientation, num_node_states,
                                      num_node_states_sq, link_state,
                                      n_xn, event_queue,
                                      next_update, xn_to, xn_rate)


@cython.boundscheck(False)
cpdef double run_cts_lean(double run_to, double current_time,
                     object event_queue,
                     np.ndarray[DTYPE_t, ndim=1] next_update,                  
                     np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_tail,                  
                     np.ndarray[DTYPE_INT_t, ndim=1] node_at_link_head,                  
                     np.ndarray[DTYPE_INT_t, ndim=1] node_state,            
                     np.ndarray[DTYPE_INT_t, ndim=1] link_state,
                     np.ndarray[DTYPE_UINT8_t, ndim=1] status_at_node,
                     np.ndarray[DTYPE_INT8_t, ndim=1] link_orientation,
                     np.ndarray[DTYPE_INT_t, ndim=1] n_xn,
                     np.ndarray[DTYPE_INT_t, ndim=2] xn_to,
                     np.ndarray[DTYPE_t, ndim=2] xn_rate, 
                     np.ndarray[DTYPE_INT_t, ndim=2] links_at_node,
                     np.ndarray[DTYPE_INT8_t, ndim=2] active_link_dirs_at_node,
                     DTYPE_INT_t num_node_states,
                     DTYPE_INT_t num_node_states_sq,
                     np.ndarray[DTYPE_INT8_t, ndim=1] bnd_lnk):
    """Run the model forward for a specified period of time. This "lean"
    version omits parameters related to property exchange and callback fn.

    Parameters
    ----------
    run_to : float
        Time to run to, starting from self.current_time
    (see celllab_cts.py for other parameters)
    """
    cdef Event ev
    
    # Continue until we've run out of either time or events
    while current_time < run_to and event_queue:

        # Is there an event scheduled to occur within this run?
        if event_queue[0].time <= run_to:

            #print 'popping'
            
            # If so, pick the next transition event from the event queue
            ev = heappop(event_queue)

            # ... and execute the transition
            do_transition_lean(ev, next_update,
                              node_at_link_tail,
                              node_at_link_head,
                              node_state, link_state,
                              status_at_node, link_orientation,
                              n_xn, xn_to, xn_rate,
                              links_at_node,
                              active_link_dirs_at_node,
                              num_node_states, num_node_states_sq,
                              bnd_lnk, event_queue)

            # Update current time
            current_time = ev.time
            
        # If there is no event scheduled for this span of time, simply
        # advance current_time to the end of the current run period.
        else:
            current_time = run_to

    return current_time



import numpy as np

from ...core.utils import as_id_array


def get_links(nodes_at_link, sort=True):
    """Get links and their directions at each node.

    Parameters
    ----------
    nodes_at_link : ndarray of int, shape `(n_links, 2)`
        Node identifier for each link tail and head.

    Returns
    -------
    tuple of ndarray of int
        Tuple of link identifiers for each node, link directions for each
        node and offsets into these arrays for each node.

    Examples
    --------
    ::
        0 -- 1
        |    | \
        2 -- 3 - 4

    >>> import numpy as np
    >>> nodes_at_link = np.array([[2, 3], [3, 4], [2, 0], [3, 1], [4, 1],
    ...                           [0, 1]])
    >>> (links_at_node, link_dirs_at_node, offset_to_node) = get_links(nodes_at_link)
    >>> links_at_node
    array([2, 5, 3, 4, 5, 0, 2, 0, 1, 3, 1, 4])
    >>> link_dirs_at_node
    array([ 1, -1,  1,  1,  1, -1, -1,  1, -1, -1,  1, -1])
    >>> offset_to_node
    array([ 0,  2,  5,  7, 10, 12])
    """
    sorted = np.argsort(nodes_at_link.reshape((-1, )))

    links_at_node = sorted // 2
    link_dirs_at_node = sorted % 2
    link_dirs_at_node[link_dirs_at_node == 0] = -1

    (_, links_per_node) = np.unique(nodes_at_link, return_counts=True)

    offset_to_node = np.empty(len(links_per_node) + 1, dtype=int)
    offset_to_node[0] = 0
    links_per_node.cumsum(out=offset_to_node[1:])

    return (as_id_array(links_at_node), as_id_array(link_dirs_at_node),
            offset_to_node)



# -*- coding: utf-8 -*-
"""
Created on Fri Apr  8 08:32:48 2016

@author: RCGlade
"""

from landlab import Component
import numpy as np
from landlab import INACTIVE_LINK

class DepthDependentDiffuser(Component):

    """
    This component implements a depth and slope dependent linear diffusion rule
    in the style of Johnstone and Hilley (2014).

    This component will ignore soil thickness located at non-core nodes.

    Parameters
    ----------
    grid: ModelGrid
        Landlab ModelGrid object
    linear_diffusivity: float
        Hillslope diffusivity, m**2/yr
        Equivalent to the soil creep efficiency
        times the soil transport decay depth.
    soil_transport_decay_depth: float
        characteristic transport soil depth, m

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import ExponentialWeatherer
    >>> from landlab.components import DepthDependentDiffuser
    >>> mg = RasterModelGrid((5, 5))
    >>> soilTh = mg.add_zeros('node', 'soil__depth')
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> BRz = mg.add_zeros('node', 'bedrock__elevation')
    >>> expweath = ExponentialWeatherer(mg)
    >>> DDdiff = DepthDependentDiffuser(mg)
    >>> expweath.calc_soil_prod_rate()
    >>> np.allclose(mg.at_node['soil_production__rate'][mg.core_nodes], 1.)
    True
    >>> DDdiff.soilflux(2.)
    >>> np.allclose(mg.at_node['topographic__elevation'][mg.core_nodes], 0.)
    True
    >>> np.allclose(mg.at_node['bedrock__elevation'][mg.core_nodes], -2.)
    True
    >>> np.allclose(mg.at_node['soil__depth'][mg.core_nodes], 2.)
    True

    Now with a slope:

    >>> mg = RasterModelGrid((3, 5))
    >>> soilTh = mg.add_zeros('node', 'soil__depth')
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> BRz = mg.add_zeros('node', 'bedrock__elevation')
    >>> z += mg.node_x.copy()
    >>> BRz += mg.node_x/2.
    >>> soilTh[:] = z - BRz
    >>> expweath = ExponentialWeatherer(mg)
    >>> DDdiff = DepthDependentDiffuser(mg)
    >>> expweath.calc_soil_prod_rate()
    >>> np.allclose(
    ...     mg.at_node['soil_production__rate'][mg.core_nodes],
    ...     np.array([ 0.60653066,  0.36787944,  0.22313016]))
    True
    >>> DDdiff.soilflux(2.)
    >>> np.allclose(
    ...     mg.at_node['topographic__elevation'][mg.core_nodes],
    ...     np.array([ 1.47730244,  2.28949856,  3.17558975]))
    True
    >>> np.allclose(mg.at_node['bedrock__elevation'][mg.core_nodes],
    ...     np.array([-0.71306132,  0.26424112,  1.05373968]))
    True
    >>> np.allclose(mg.at_node['soil__depth'], z - BRz)
    True
    """

    _name = 'DepthDependentDiffuser'

    _input_var_names = (
        'topographic__elevation',
        'soil__depth',
        'soil_production__rate',
    )

    _output_var_names = (
        'soil__flux',
        'topographic__slope',
        'topographic__elevation',
        'bedrock__elevation',
        'soil__depth',
    )

    _var_units = {
        'topographic__elevation' : 'm',
        'topographic__slope' : 'm/m',
        'soil__depth' : 'm',
        'soil__flux' : 'm^2/yr',
        'soil_production__rate' : 'm/yr',
        'bedrock__elevation' : 'm',
    }

    _var_mapping = {
        'topographic__elevation' : 'node',
        'topographic__slope' : 'link',
        'soil__depth' : 'node',
        'soil__flux' : 'link',
        'soil_production__rate' : 'node',
        'bedrock__elevation' :'node',
    }

    _var_doc = {
        'topographic__elevation':
                'elevation of the ground surface',
        'topographic__slope':
                'gradient of the ground surface',
        'soil__depth':
                'depth of soil/weather bedrock',
        'soil__flux':
                'flux of soil in direction of link',
        'soil_production__rate':
                'rate of soil production at nodes',
        'bedrock__elevation':
                'elevation of the bedrock surface',
    }
    def __init__(self,grid, linear_diffusivity=1.0,
                 soil_transport_decay_depth=1.0):
        """Initialize the DepthDependentDiffuser."""

        # Store grid and parameters
        self._grid = grid
        self.K = linear_diffusivity
        self.soil_transport_decay_depth = soil_transport_decay_depth

        # create fields
        # elevation
        if 'topographic__elevation' in self.grid.at_node:
            self.elev = self.grid.at_node['topographic__elevation']
        else:
            self.elev = self.grid.add_zeros('node', 'topographic__elevation')

        # slope
        if 'topographic__slope' in self.grid.at_link:
            self.slope = self.grid.at_link['topographic__slope']
        else:
            self.slope = self.grid.add_zeros('link', 'topographic__slope')

        # soil depth
        if 'soil__depth' in self.grid.at_node:
            self.depth = self.grid.at_node['soil__depth']
        else:
            self.depth = self.grid.add_zeros('node', 'soil__depth')

        # soil flux
        if 'soil__flux' in self.grid.at_link:
            self.flux = self.grid.at_link['soil__flux']
        else:
            self.flux=self.grid.add_zeros('link', 'soil__flux')

        # weathering rate
        if 'soil_production__rate' in self.grid.at_node:
            self.soil_prod_rate = self.grid.at_node['soil_production__rate']
        else:
            self.soil_prod_rate = self.grid.add_zeros('node',
                                                      'soil_production__rate')

        # bedrock elevation
        if 'bedrock__elevation' in self.grid.at_node:
            self.bedrock = self.grid.at_node['bedrock__elevation']
        else:
            self.bedrock = self.grid.add_zeros('node', 'bedrock__elevation')

    def soilflux(self, dt):
        """Calculate soil flux for a time period 'dt'.

        Parameters
        ----------

        dt: float (time)
            The imposed timestep.
        """

        #update soil thickness
        self.grid.at_node['soil__depth'][:] = (
            self.grid.at_node['topographic__elevation']
            - self.grid.at_node['bedrock__elevation'])

        #Calculate soil depth at links.
        H_link = self.grid.map_value_at_max_node_to_link(
            'topographic__elevation','soil__depth')

        #Calculate gradients
        slope = self.grid.calc_grad_at_link(self.elev)
        slope[self.grid.status_at_link == INACTIVE_LINK] = 0.

        #Calculate flux
        self.flux[:] = (-self.K
                        * slope
                        * (1.0 - np.exp(-H_link
                                        / self.soil_transport_decay_depth)))

        #Calculate flux divergence
        dqdx = self.grid.calc_flux_div_at_node(self.flux)

        #Calculate change in soil depth
        dhdt = self.soil_prod_rate - dqdx

        #Calculate soil depth at nodes
        self.depth[self.grid.core_nodes] += dhdt[self.grid.core_nodes] * dt

        #prevent negative soil thickness
        self.depth[self.depth < 0.0] = 0.0

        #Calculate bedrock elevation
        self.bedrock[self.grid.core_nodes] -= (
            self.soil_prod_rate[self.grid.core_nodes] * dt)

        #Update topography
        self.elev[self.grid.core_nodes] = (self.depth[self.grid.core_nodes]
                                         + self.bedrock[self.grid.core_nodes])

    def run_one_step(self, dt):
        """

        Parameters
        ----------
        dt: float (time)
            The imposed timestep.
        """

        self.soilflux(dt)
